#!/usr/bin/python
#  Copyright (c) 2013, Intel Corporation. All rights reserved.

class DEFINES( object ):
  """ Defines Class"""

  def __init__( self ):
    pass

  LCP_VERSION = 2
  TOOL_VERSION = '0.5'
  BUILD_DATE = '2013-09-06'

  SUPPORTED_LCP_VERSION = {
    # Policy Version : List Verson
    '3.1'            : '2.1',
    '3.0'            : '2.0'
  }

  # Policy Type global values:
  LIST = 0
  ANY = 1

  # Policy Rules global values
  PsRules = 0
  PoRules = 1

  # max value for Min SINIT Version and BIOS & SINIT Revocation Limit fields
  maxVersion = 255

  # The PolicyControlField provides a number of control bits which are defined as:
  #   Bit31 AUXDeletionControl
  #   Bits30:4 Reserved and should set to zero
  #   Bit20 Ignore PS STM
  #   Bit17 Ignore PS PCONF
  #   Bit16 Ignore PS MLE
  #   Bit 3 Indicates Force Owner Policy
  #   Bit 2 Allow NPW.
  #   Bit 1 PCR 17
  #   Bit 0 Reserved and should set to zero
  #
  PolicyControlAUXDeletionControl   = 0x80000000   # bit 31
  PolicyControlIgnorePsStmBitMask   = 0x00100000   # bit 20
  PolicyControlIgnorePsPconfBitMask = 0x00020000   # bit 17
  PolicyControlIgnorePsMleBitMask   = 0x00010000   # bit 16
  PolicyControlForceOwnerBitMask    = 0x00000008   # bit 3
  PolicyControlPcr17BitMask         = 0x00000004   # bit 2
  PolicyControlAllowNpwBitMask      = 0x00000002   # bit 1

  # List Policy Signing Algorithm
  LCP_POLSALG_NONE        = 0   # list not signed
  LCP_POLSALG_RSA_PKCS_15 = 1   # list signed
  LCP_POLSALG_ECDSA       = 2
  LCP_POLSALG_SM2         = 3

  # Element types:  MLE, PCONF, SBIOS and STM
  LCP_POLELT_TYPE_MLE    = 0x00
  LCP_POLELT_TYPE_PCONF  = 0x01
  LCP_POLELT_TYPE_SBIOS  = 0x02
  LCP_POLELT_TYPE_MLE2   = 0x10
  LCP_POLELT_TYPE_PCONF2 = 0x11
  LCP_POLELT_TYPE_SBIOS2 = 0x12
  LCP_POLELT_TYPE_STM2   = 0x14

  # Each list can have each element with each possible hash algorithm
  # So Pdef_list.[Mle,Pconf,Sbios,Stm]DefData[x]  is now an array
  # where x is one of the indexes below which must be < MAX_ELEMENTS
  DEFDATA_INDEX = {
    'SHA1'      : 0,
    'SHA256'    : 1,
    'SHA384'    : 2,
    'SHA512'    : 3,
    #'SM3'       : 4
  }


  # Other than ELEMENT_NAME_NONE with no '-' character,
  # all other names must be <element>-<hash>
  # This is used in list.py to create new element of the hash type.
  # So the following element names are possible:
  ELEMENT_NAME_NONE          = "None"
  ELEMENT_NAME_MLE_SHA1      = "MLE-SHA1"
  ELEMENT_NAME_MLE_SHA256    = "MLE-SHA256"
  ELEMENT_NAME_MLE_SHA384   = "MLE-SHA384"
  ELEMENT_NAME_MLE_SHA512   = "MLE-SHA512"
  ELEMENT_NAME_MLE_SM3       = "MLE-SM3"

  ELEMENT_NAME_PCONF_SHA1    = "PCONF-SHA1"
  ELEMENT_NAME_PCONF_SHA256  = "PCONF-SHA256"
  ELEMENT_NAME_PCONF_SHA384 = "PCONF-SHA384"
  ELEMENT_NAME_PCONF_SHA512 = "PCONF-SHA512"
  ELEMENT_NAME_PCONF_SM3     = "PCONF-SM3"

  ELEMENT_NAME_SBIOS_SHA1    = "SBIOS-SHA1"
  ELEMENT_NAME_SBIOS_SHA256  = "SBIOS-SHA256"
  ELEMENT_NAME_SBIOS_SHA384 = "SBIOS-SHA384"
  ELEMENT_NAME_SBIOS_SHA512 = "SBIOS-SHA512"
  ELEMENT_NAME_SBIOS_SM3     = "SBIOS-SM3"

  ELEMENT_NAME_STM_SHA1      = "STM-SHA1"
  ELEMENT_NAME_STM_SHA256    = "STM-SHA256"
  ELEMENT_NAME_STM_SHA384   = "STM-SHA384"
  ELEMENT_NAME_STM_SHA512   = "STM-SHA512"
  ELEMENT_NAME_STM_SM3       = "STM-SM3"

  ELEMENT_NAME_MLE_LEGACY    = "MLE-LEGACY"
  ELEMENT_NAME_PCONF_LEGACY  = "PCONF-LEGACY"
  ELEMENT_NAME_SBIOS_LEGACY  = "SBIOS-LEGACY"

  # element name strings for PO rules
  ELEMENT_PO_RULES = [
    ELEMENT_NAME_STM_SHA512,   ELEMENT_NAME_STM_SHA384,
    ELEMENT_NAME_STM_SHA256,   ELEMENT_NAME_STM_SHA1,     #ELEMENT_NAME_STM_SM3,
    ELEMENT_NAME_PCONF_SHA512, ELEMENT_NAME_PCONF_SHA384,
    ELEMENT_NAME_PCONF_SHA256, ELEMENT_NAME_PCONF_SHA1,   #ELEMENT_NAME_PCONF_SM3,
    ELEMENT_NAME_PCONF_LEGACY,
    ELEMENT_NAME_MLE_SHA512,   ELEMENT_NAME_MLE_SHA384,
    ELEMENT_NAME_MLE_SHA256,   ELEMENT_NAME_MLE_SHA1,     #ELEMENT_NAME_MLE_SM3,
    ELEMENT_NAME_MLE_LEGACY
  ]

  """getElement - return array of element names strings"""
  ELEMENT = [
    ELEMENT_NAME_SBIOS_SHA512, ELEMENT_NAME_SBIOS_SHA384,
    ELEMENT_NAME_SBIOS_SHA256, ELEMENT_NAME_SBIOS_SHA1,   #ELEMENT_NAME_SBIOS_SM3,
    ELEMENT_NAME_SBIOS_LEGACY,
    ELEMENT_NAME_STM_SHA512,   ELEMENT_NAME_STM_SHA384,
    ELEMENT_NAME_STM_SHA256,   ELEMENT_NAME_STM_SHA1,     #ELEMENT_NAME_STM_SM3,
    ELEMENT_NAME_PCONF_SHA512, ELEMENT_NAME_PCONF_SHA384,
    ELEMENT_NAME_PCONF_SHA256, ELEMENT_NAME_PCONF_SHA1,   #ELEMENT_NAME_PCONF_SM3,
    ELEMENT_NAME_PCONF_LEGACY,
    ELEMENT_NAME_MLE_SHA512,   ELEMENT_NAME_MLE_SHA384,
    ELEMENT_NAME_MLE_SHA256,   ELEMENT_NAME_MLE_SHA1,     #ELEMENT_NAME_MLE_SM3,
    ELEMENT_NAME_MLE_LEGACY
  ]
  
  # This replaces util.getHashes()
  # supported hash algorithm names
  SUPPORTED_HASHES = [
    'SHA1', 'SHA256', 'SHA384', 'SHA512',
    #'SM3'
  ]
  
  # allowed hash algorithm names
  ALLOWED_HASHES = [
    'SHA1',
    #'SHA224', # not supported
    'SHA256', 'SHA384', 'SHA512',
    'SM3'
  ]
  
  # ALLOWED_SIGNATURE_SCHEMES is ordered list that references TPM_ALG_SIGN_MASK
  # allowed signing algorithm names
  ALLOWED_SIGNATURE_SCHEMES = [
    #'RSA-1024-SHA1', 'RSA-1024-SHA256', 'RSA-2048-SHA1',     # not supported
    'RSA-2048-SHA256',
    #'RSA-2048-SHA384', 'RSA-2048-SHA512',                    # not supported
    #'RSA-3072-SHA256', 'RSA-3072-SHA384', 'RSA-3072-SHA512', # not supported
    #'RSA-4096-SHA256', 'RSA-4096-SHA384', 'RSA-4096-SHA512', # not supported
    'ECDSA P-256', 'ECDSA P-384',
    'SM2'
  ]
  
  SIGNATURE_ALGORITHMS = [
    'None',
    #'RSA PKCS1.5/SHA1',
    'RSA PKCS1.5/SHA256',
    #'RSA PKCS1.5/SHA384',
    #'RSA PKCS1.5/SHA512',
    'ECDSA P-256/SHA256',
    'ECDSA P-384/SHA384',
    #'SM2/SM3'
  ]
  
  SIGNATURE_KEY_SIZE = {
    'None'                : [],
    'RSA PKCS1.5/SHA256'  : ['2048'],
    'ECDSA P-256/SHA256'  : ['256'],
    'ECDSA P-384/SHA384'  : ['384'],
    'SM2/SM3'             : ['256']
  }

  # Hash Algorithm defines for pdef.hashAlg
  TPM_ALG_HASH = {
    'SHA1_LEGACY'         : 0x0000,
    'SHA1'                : 0x0004,
    'SHA256'              : 0x000B,
    'SHA384'              : 0x000C,
    'SHA512'              : 0x000D,
    'NULL'                : 0x0010,
    'SM3'                 : 0x0012
  }

  # Signature Algorithm defined for LCP_POLICY_LIST2.SigAlgorithm
  TPM_ALG_SIGN = {
    'NULL'                : 0x0010,  # same as in Hash Algorithm
    'RSASSA'              : 0x0014,
    'ECDSA'               : 0x0018,
    'SM2'                 : 0x001B
  }

  # SHAXXXXX_DIGEST_SIZE
  DIGEST_SIZE = {
    'SHA1'                : 20,
    'SHA256'              : 32,
    'SHA384'              : 48,
    'SHA512'              : 64,
    'SM3'                 : 32
  }

  # TPM_ALG_HASH_MASK_XXXX
  TPM_ALG_HASH_MASK = {
    'SHA1'                : 0x0001,
    'SHA224'              : 0x0002,
    'SHA512_224'          : 0x0004,
    'SHA256'              : 0x0008,
    'SHA512_256'          : 0x0010,
    'SM3'                 : 0x0020,
    'SHA384'              : 0x0040,
    'SHA512'              : 0x0080,
    'WHIRLPOOL'           : 0x0100
  }

  # TPM_ALG_SIGN_MASK_XXXX
  TPM_ALG_SIGN_MASK = {
    'RSA-1024-SHA1'       : 0x00000001,
    'RSA-1024-SHA256'     : 0x00000002,
    'RSA-2048-SHA1'       : 0x00000004,
    'RSA-2048-SHA256'     : 0x00000008,
    'RSA-2048-SHA384'     : 0x00000010,
    'RSA-2048-SHA512'     : 0x00000020,
    'RSA-3072-SHA256'     : 0x00000040,
    'RSA-3072-SHA384'     : 0x00000080,
    'RSA-3072-SHA512'     : 0x00000100,
    'RSA-4096-SHA256'     : 0x00000200,
    'RSA-4096-SHA384'     : 0x00000400,
    'RSA-4096-SHA512'     : 0x00000800,
    'ECDSA P-256'         : 0x00001000,
    'ECDSA P-384'         : 0x00002000,
    'SM2'                 : 0x00010000
  }

  # Types of hash files - see util.verifyHashFile()

  HashFileMode = {
    'HdrNull'     : 0,
    'HdrSHA1'     : 1,
    'RawSHA1'     : 2,
    'RawSHA256'   : 3,
    'RawSHA384'   : 4,
    'RawSHA512'   : 5,
    'RawSM3'      : 6 #TODO: check the value
  }

  # Types of PCR dump Files - see util.verifyPcrFile()
  PcrFileMode = {
    'Null'         : 0,
    'Pcrd'         : 7,
    'Pcr2'         : 8
  }

  # key files are private or public keys for RSA or ECC
  KEY_FILE_TYPE = {
    'PRIVATE_RSASSA': 1,
    'PUBLIC_RSASSA' : 2,
    'PRIVATE_ECDSA' : 3,
    'PUBLIC_ECDSA'  : 4
  }

  PCRDFileHdrSize         = 16
  PCR2FileHdrSize         = 8
  PCRFileMinHdrSize       = 7   # 1st 6 bytes of PCRD or PCR2 file are enough to determine the type and check the HashAlg






#!/usr/bin/python
#  Copyright (c) 2013, Intel Corporation. All rights reserved.

# using print() built infunction, disable print statement
from __future__ import print_function

import os
import time

from defines import DEFINES
from LcpPolicy import *
from util import UTILS
utilities = UTILS()


_GlobalHashData   = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]

PCONF_ELEMENT2_HDR_SIZE = 16  # size of LCP_PCONF_ELEMENT2's fields except PcrInfo's
                              # i.e. sizeof(ElementSize+ElementType+
                              #      PolEltControl+HashAlg+NumPCRInfos)

PCR_INFO2_HDR_SIZE = 12       # size of TPMS_QUOTE_INFO's fileds except buffer
                              # i.e. sizeof(count+hash+sizeOfSelect+pcrSelect+size)

SHA1_PCR_INFO_SIZE = 26       # for PconfLegacyElement


# TXT Policy Generator Tool
# PDEF Class - Policy Definition File Structure
#
class PDEF( object ):
  """ PDEF Class"""

  def __init__( self ):
    """__init__() - PDEF class constructor"""
    self.WorkingDirectory = ""
    self.FileTypeSignature  = "TXT Policy Definition v2"       # CHAR8 array[24]
    self.DefCompany         = "Intel"                          # CHAR8 array[16]
    self.StructVersion      = 2                                # incremented for any change to structure
    #
    # Structure capacity - can change without incrementing StructVersion
    #
    self.MaxLists =          8                                 # UINT8  MAX_LISTS = 8      - Max lists per policy
    #self.MaxElements =       4                                 # UINT8  MAX_ELEMENTS = 4   - Max elements per list per element
    self.MaxElements =       2                                 # only support SHA1 & SHA256 so far, not yet SHA384 & SHA512

    self.MaxHashSize =       64                                # UINT8  MAX_HASH_SIZE = 64 - Max allowed HASH size
    self.MaxHashes =         16                                # UINT16 MAX_HASHES = 16    - Max hashes per element, 0=unlimited
    self.MaxFileNameSize =   32                                # UINT8  MAX_FILENAME_SIZE 32 - Max size for a filename
    # ReservedCap[6]    0,                                     # UINT8                       - reserved for future definition
    #
    # Start of variable data whose content can be changed by the tool
    #
    self.ToolDate =          time.strftime("%Y%m%d")           # UINT32   YYYYMMDD - Build date of the tool
    #self.ToolVersion =      0x0200                            # UINT16   Version of the tool
    self.ToolVersionMajor =  02
    self.ToolVersionMinor =  00
    self.Rules =             1                                 # UINT8    Type of rules: 0=PS, 1=PO
    self.Modified =          0                                 # BOOLEAN, True =changed since last 'BUILD'
    # ReservedTool                                             # UINT16
    #
    # Start of Policy Definition - same as NV Policy Structure
    #
    #self.PolVersion =        0x0301                           # UINT16
    self.PolVersionMajor =    03
    self.PolVersionMinor =    01
    self.HashAlg  =           DEFINES.TPM_ALG_HASH['SHA256']   # UINT16  TPM_ALG_XXXX
    self.PolicyType =         1                                # UINT8   0=LIST, 1=ANY
    self.SinitMinVersion =    0                                # UINT8
    self.DataRevocationCounters = [0,0,0,0,0,0,0,0]            # UINT16  DataRevocationCounters[MAX_LISTS] Default is 0's
    self.PolicyControl =      0                                # UINT32  Encoding of (NPW, PCR17, Force PO)
    self.MaxSinitMinVersion = 255                              # UINT8   reserved in PS Policy
    self.MaxBiosMinVersion =  255                              # UINT8   reserved in PS Policy
    # HashAlg is already defined above
    # LcpHashAlgMask should be the content of CheckListBox
    self.LcpHashAlgMask = DEFINES.TPM_ALG_HASH_MASK['SHA256']  # UINT16  TPM_ALG_HASH_MASK_XXXXX
    # Should AuxHashAlgMask be renamed to ApAlg?
    self.AuxHashAlgMask = DEFINES.TPM_ALG_HASH_MASK['SHA256']  # UINT16  TPM_ALG_HASH_MASK_XXXXX
    self.LcpSignAlgMask = DEFINES.TPM_ALG_HASH_MASK['SHA256']  # UINT32  TPM_ALG_SIGN_MASK_XXXXX

    # save both the raw and hex formatted versions of the SHA1 and SHA256 hashes
    #self.PolicyHashSha1       = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    #                             0,0,0,0]                          # MAX_HASH = 20 for SHA1,
    #self.PolicyHashSha1Hex    = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    #                             0,0,0,0]
    #self.PolicyHashSha256     = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    #                             0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]  # MAX_HASH = 32 for SHA256
    #self.PolicyHashSha256Hex  = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    #                             0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
    # For Python, don't need multiple variable of different size to store hash values.
    # Going forward, there will be more hash algorithms.
    self.PolicyHash           = []
    self.PolicyHashHex        = []

    #
    # Start of Policy Data Structure Definition
    #
    #self.LastBuildDateStamp = 20000101                        # UINT32 - YYYYMMDD - date of last build
    self.LastBuildDateStampYear  = 2000                        #    UINT16 - YYYY  - year of last build
    self.LastBuildDateStampMonth = 01                          #    UINT8  - MM    - Month of last build
    self.LastBuildDateStampDay   = 01                          #    UINT8  - DD    - Day of last build
    #self.LastBuildTimeStamp = 00000000                        # UINT32 - HHMMSS00 - time of last build
    self.LastBuildTimeStampHour   = 00                         #    UINT8 - HH     - hour of last build
    self.LastBuildTimeStampMinute = 00                         #    UINT8 - MM     - minute of last build
    self.LastBuildTimeStampSecond = 00                         #    UINT8 - SS     - second of last build
    self.LastBuildTimeStampLowByte = 00                        #    UINT8 - 00     - read only
    # Reserved3                                                # UINT16
    self.CurrentListView =    0                                # UINT8    - saves user state, indicates list  to display
    self.NumLists =           0                                # UINT8    - Actual number of lists
    # PLIST_DEF PolListInfo[0-7]                               # PLIST_DEF - dictionary containing policy is populated in addPlistDef
    self.PolListInfo = {'0':None, '1':None,
                        '2':None, '3':None,
                        '4':None, '5':None,
                        '6':None, '7':None}

  #
  # add PolListInfo[0-7] to the PDEF
  #
  def addPlistDef(self, listNumber):
    """ addPlistDef - add a PLIST_DEF to the PDEF"""
    print("PLIST_DEF::addPlistDef - list %i" % (listNumber))

    if(listNumber > self.MaxLists):           # should never get here if caller checked MAX_LISTS correctly ...
      print("PLIST_DEF::addPlistDef - list %i > MAX=%i" % (listNumber, self.MaxLists))
    else:
      self.PolListInfo[str(listNumber-1)] = PLIST_DEF()
      #self.PolListInfo[str(listNumber-1)].append(PLIST_DEF())
      self.PolListInfo[str(listNumber-1)].ListValid = True
      #print("addPlistDef %d = %s" % (listNumber, self.PolListInfo[str(listNumber-1)])) # DBGDBG

      # at the time when adding the list to policy, sync policy version number to list version number
      policyversion = str(self.PolVersionMajor)+'.'+str(self.PolVersionMinor)
      listversion = DEFINES.SUPPORTED_LCP_VERSION[policyversion]
      majorstring, minorstring = listversion.split('.')
      self.PolListInfo[str(listNumber-1)].ListVersionMajor = int(majorstring)
      self.PolListInfo[str(listNumber-1)].ListVersionMinor = int(minorstring)


  def getCurrentListObject(self):
    """getCurrentListObject - return the current list object per CurrentListView"""
    listNumber = self.CurrentListView
    #print("getCurrentListObject - CurrentListView = %i" % (listNumber))

    if(listNumber > self.MaxLists):           # should never get here if CurrentListView was set correctly ...
      print("PLIST_DEF::getCurrentListObject - list %i > MAX_LISTS" % (listNumber))
    elif(listNumber <= 0):
      return self.PolListInfo[str(0)]
    else:
      return self.PolListInfo[str(listNumber-1)]


class PLIST_DEF( object ):
  """PLIST_DEF class"""

  def __init__(self):
    """___init__() = PLIST_DEF class constructor"""
    #print("constructing a PLIST_DEF")

    self.LdefSize                = 128                     # UINT32 - number of bytes in this LIST def
    self.Tag                     = "LIST"                  # UINT32 - confirms this is a LIST def struct
    #self.ListVersion             = 0x0201                 # UINT16 - version 2.1
    self.ListVersionMajor        = 02
    self.ListVersionMinor        = 01
    self.ListValid               = False                   # BOOLEAN - indicates if this list is to be included
    self.ListModified            = False                   # BOOLEAN - indicates if this list has changed since last build; only clear when building policy.
    # Reserved[3]           = 0                            # UINT8
    self.SigAlgorithm            = DEFINES.TPM_ALG_SIGN['NULL']    # UINT8 - USER: 0=Not signed, 1 = PSA PKCS15
    self.sigAlgorithmHash        = DEFINES.TPM_ALG_HASH['SHA1']    # corresponds to 1
    self.PolicyElementSize       = 0                       # UINT32 - total size of all elements in this list
    self.CurrentElementView      = "None"                  # UINT8 - saves user state: 0=MLE, 1=PCONF, 3=SBIOS, 0xFF=None
    # Reserved2[3]          = 0                            # UINT8

    # start of signature
    self.SyncRevCount            = True                    # BOOLEAN - if true, updates Policy with RevocationCounter value
    # Reserved3[3]          = 0                            # UINT8
    self.RevokeCounter           = 0                       # UINT16 - value to be populated in PDEF.RevocationCounter[n]
    self.RevocationCounter       = 0                       # UINT16
    self.KeySize                 = 2048                    # UINT16 -  1024, 2048 or 3072, default = 2048
    self.PubKeyFile              = ""                      # CHAR16[MAX_FILENAME_SIZE=32] - filename of signing key
    self.PvtKeyFile              = ""
    self.PubKeyData              = ""                      # PubKeyData[KeySize] for RSA - as a binary string
    # Not creating a separate class because the GUI will create and destroy when selecting between RSA and ECC signature algs
    self.PubKeyQx                = ""                      # PubKeyData[KeySize] for ECC x coord - as binary string
    self.PubKeyQy                = ""                      # PubKeyData[KeySize] for ECC y coord - as binary string
    self.PvtKeyFileIsSignature   = False                   # False: PvtKeyFile is a key file; True: PvtKeyFile is signature file

    # using one collection(list) to store all element definition data
    # Python list guarantees the order.
    self.ElementDefData = []


# MLE Element definition - TYPE=0
class MLE_DEF( object ):
  """MLE_DEF class"""

  def __init__(self, hashAlg):
    """__init__() = MLE_DEF class constructor"""
    #print("constructing a MLE_DEF")

    try:
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if hashAlg == val).next()
    except StopIteration:
      hashAlgname = ""
      print("MLE_DEF::__init__ - invalid hashAlg=%d" % (hashAlg))

    self.Name                = "MLE-"+hashAlgName          # String - Name used for GUI identification
    self.InfoSize            = 24                          # UINT32 - number of bytes in this structure
    self.Tag                 = "MLE_"                      # UINT32 - confirms this is a MLE_DEF
    self.IncludeInList       = False                       # BOOLEAN - indicates this LIST includes an MLE element
    #Reserved3[5]            = 0                           # UINT8
    self.SinitMinVersion     = 0                           # UINT8 - USER
    self.HashAlg             = hashAlg                     # UINT16 - TPM_ALG_SHAXXXXX
    self.Control             = 0                           # UINT32 - USER: Bit0: Ignore PS MLE elements
    self.NumbHashes          = 0                           # UINT16 - Tracks number of valid entries in MleHashFiles[]
    self.CurrentView         = 0                           # UINT16 - User state: last MleHashFiles[] selected
    self.HashFiles           = []                          # variable size array containing filenames of hashes

  # Build the MLE element and return its size
  # thisPdefList - is the list's source data from pdef.PolListInfo[list]
  # policyElement - is the destination LCP POLICY element
  # Return the elementSize built, or 0 if an error occurs
  #
  def build(self, thisPdefList, policyElements, cwd):
    """buildMleElement - build the MLE element"""
    func = 'buildMleElement'

    print("%s" %(func )) # DBGDBG
    elementSize = 18        # size all fields except Hashes[] in bytes

    # build the element's data
    #mleDefData = thisPdefList.MleDefData[index]
    policyElement = LCP_MLE_ELEMENT2()
    policyElement.PolEltControl = self.Control
    policyElement.SINITMinVersion = self.SinitMinVersion
    policyElement.HashAlg = self.HashAlg
    policyElement.NumHashes = self.NumbHashes
    print("%s - PolEltControl=%d, SINITMinVersion=%d from %d" %
          (func, policyElement.PolEltControl, policyElement.SINITMinVersion, self.SinitMinVersion)) # DBGDBG

    hashAlg = self.HashAlg
    try:
      # reverse lookup of the hash algorithm name(key) for the given HashAlg value
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == hashAlg)).next()
    except StopIteration:
      print ("HashAlg=%d is not supported, aborting build" % (hashAlg))
      return 0

    # Build the hashes from each HashFiles[]
    for file in self.HashFiles:
      hashdata = utilities.getHashFromFile(os.path.join(cwd, file), self.HashAlg)
      if(len(hashdata) != DEFINES.DIGEST_SIZE[hashAlgName]):
        #self.StatusBar.SetStatusText("Invalid hash file %s, aborting build" % (file))
        return 0

      #policyElement.Hashes += _GlobalHashData      # the hash data from the file
      policyElement.Hashes.append(hashdata)
      print("%s - policyElement.Hashes size=0x%x" % (func, len(policyElement.Hashes))) # DBGDBG

    # update elementSize
    elementSize += DEFINES.DIGEST_SIZE[hashAlgName] * self.NumbHashes

    # ElementSize = 16 +HashSize* (PDEF.PolListInfo[i].MleDefData[index].NumbHashes)
    # if HashAlg == SHA-1, HashSize=20
    policyElement.ElementSize = elementSize
    print("%s - done, size=0x%x" % (func, elementSize))

    policyElements.append(policyElement)
    return(elementSize)

  def printDef(self, f):
    """printDef - write the Mle Def to the specified human readable text file for printing"""

    #print("printMleDef - object: %s" % (mleDefData))     # DBGDBG
    print("***MLE_DEF***",                                      file=f)
    print("InfoSize",             " = ", self.InfoSize,         file=f)
    print("Tag",                  " = ", self.Tag,              file=f)
    #print("IncludeInList",        " = ", self.IncludeInList,    file=f)
    print("SinitMinVersion",      " = ", self.SinitMinVersion,  file=f)
    print("HashAlg",              " = ", hex(self.HashAlg),     file=f)
    print("Control",              " = ", self.Control,          file=f)
    print("NumbHashes",           " = ", self.NumbHashes,       file=f)
    print("CurrentView",          " = ", self.CurrentView,      file=f)
    print("HashFiles",            " = ", self.HashFiles,        file=f)


class MLELEGACY_DEF( object ):
  """MLELEGACY_DEF class"""

  def __init__(self):
    """__init__() = MLELEGACY_DEF class constructor"""
    #print("constructing a MLELEGACY_DEF")

    self.Name                = DEFINES.ELEMENT_NAME_MLE_LEGACY  # String - Name used for GUI identification
    self.InfoSize            = 24                          # UINT32 - number of bytes in this structure
    self.Tag                 = "MLE_"                      # UINT32 - confirms this is a MLE_DEF
    self.IncludeInList       = False                       # BOOLEAN - indicates this LIST includes an MLE element
    #Reserved3[5]            = 0                           # UINT8
    self.SinitMinVersion     = 0                           # UINT8 - USER
    self.HashAlg             = 0                           # UINT8 - SHA1
    self.Control             = 0                           # UINT32 - USER: Bit0: Ignore PS MLE elements
    self.NumbHashes          = 0                           # UINT16 - Tracks number of valid entries in MleHashFiles[]
    self.CurrentView         = 0                           # UINT16 - User state: last MleHashFiles[] selected
    self.HashFiles           = []                          # variable size array containing filenames of hashes

  # Build the MLE element and return its size
  # thisPdefList - is the list's source data from pdef.PolListInfo[list]
  # policyElement - is the destination LCP POLICY element
  # Return the elementSize built, or 0 if an error occurs
  #
  def build(self, thisPdefList, policyElements, cwd):
    """buildMleLegacyElement - build the MLE element"""

    func = "buildMleLegacyElement"
    print("%s" %(func)) # DBGDBG
    elementSize = 16        # size all fields except Hashes[] in bytes

    # build the element's data
    #mleDefData = thisPdefList.MleLegacyDefData[DEFINES.DEFDATA_INDEX['SHA1']]
    policyElement = LCP_MLE_ELEMENT()
    policyElement.PolEltControl = self.Control
    policyElement.SINITMinVersion = self.SinitMinVersion
    policyElement.HashAlg = self.HashAlg
    policyElement.NumHashes = self.NumbHashes
    print("%s - PolEltControl=%d, SINITMinVersion=%d from %d" %
          (func, policyElement.PolEltControl, policyElement.SINITMinVersion, self.SinitMinVersion)) # DBGDBG

    # Build the hashes from each HashFiles[]
    for file in self.HashFiles:
      hashdata = utilities.getHashFromFile(os.path.join(cwd, file), DEFINES.TPM_ALG_HASH['SHA1'])
      if(len(hashdata) != DEFINES.DIGEST_SIZE['SHA1']):
        print ("Invalid hash file %s, aborting build" % (file))
        return 0

      policyElement.Hashes.append(hashdata)      # the hash data from the file
      print("%s - policyElement.Hashes size=0x%x" % (func, len(policyElement.Hashes))) # DBGDBG

    if(self.HashAlg == DEFINES.TPM_ALG_HASH['SHA1_LEGACY']):
      # update elementSize
      elementSize += DEFINES.DIGEST_SIZE['SHA1'] * self.NumbHashes
    else:
      print ("HashAlg=%d is not supported, aborting build" % (thisPdefList.HashAlg))
      return 0

    # ElementSize = 16 +HashSize* (PDEF.PolListInfo[i].MleDefData.NumbHashes)
    # if HashAlg == SHA-1, HashSize=20
    policyElement.ElementSize = elementSize
    print("%s - done, size=0x%x" % (func, elementSize))

    policyElements.append(policyElement)
    return(elementSize)

  def printDef(self, f):
    """printMleDef - write the Mle Def to the specified human readable text file for printing"""

    #print("printMleDef - object: %s" % (mleDefData))     # DBGDBG

    print("***MLELEGACY_DEF***",                                file=f)
    print("InfoSize",             " = ", self.InfoSize,         file=f)
    print("Tag",                  " = ", self.Tag,              file=f)
    #print("IncludeInList",        " = ", self.IncludeInList,    file=f)
    print("SinitMinVersion",      " = ", self.SinitMinVersion,  file=f)
    print("HashAlg",              " = ", hex(self.HashAlg),     file=f)
    print("Control",              " = ", self.Control,          file=f)
    print("NumbHashes",           " = ", self.NumbHashes,       file=f)
    print("CurrentView",          " = ", self.CurrentView,      file=f)
    print("HashFiles",            " = ", self.HashFiles,        file=f)


# STM Element definition - TYPE=0
class STM_DEF( object ):
  """STM_DEF class"""

  def __init__(self, hashAlg):
    """__init__() = STM_DEF class constructor"""
    #print("constructing a STM_DEF")

    try:
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if hashAlg == val).next()
    except StopIteration:
      hashAlgname = ""
      print("STM_DEF::__init__ - invalid hashAlg=%d" % (hashAlg))

    self.Name                = "STM-"+hashAlgName          # String - Name used for GUI identification
    self.InfoSize            = 24                          # UINT32 - number of bytes in this structure
    self.Tag                 = "STM_"                      # UINT32 - confirms this is a MLE_DEF
    self.IncludeInList       = False                       # BOOLEAN - indicates this LIST includes an MLE element
    #Reserved3[5]            = 0                           # UINT8
    self.HashAlg             = hashAlg                     # UINT16 - TPM_ALG_SHAXXXXX
    self.Control             = 0                           # UINT32 - USER: Bit0: Ignore PS MLE elements
    self.NumbHashes          = 0                           # UINT16 - Tracks number of valid entries in MleHashFiles[]
    self.CurrentView         = 0                           # UINT16 - User state: last MleHashFiles[] selected
    self.HashFiles           = []

  def build(self, thisPdefList, policyElements, cwd):
    """buildStmElement - build the STM element"""

    func='buildStmElement'
    print("%s" % (func)) # DBGDBG
    elementSize = 16        # size all fieds except Hashes[] in bytes

    # build the element's data
    #stmDefData = thisPdefList.StmDefData[index]
    policyElement = LCP_STM_ELEMENT2()
    policyElement.PolEltControl = self.Control
    policyElement.HashAlg = self.HashAlg
    policyElement.NumHashes = self.NumbHashes
    print("%s - PolEltControl=%d" % (func, policyElement.PolEltControl)) # DBGDBG


    hashAlg = self.HashAlg
    try:
      # reverse lookup of the hash algorithm name(key) for the given HashAlg value
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == hashAlg)).next()
    except StopIteration:
      print ("HashAlg=%d is not supported, aborting build" % (hashAlg))
      return 0

    # Build the hashes from each HashFiles[]
    for file in self.HashFiles:
      hashdata = utilities.getHashFromFile(os.path.join(cwd, file), self.HashAlg)
      if(len(hashdata) != DEFINES.DIGEST_SIZE[hashAlgName]):
        #self.StatusBar.SetStatusText("Invalid hash file %s, aborting build" % (file))
        return 0

      #policyElement.Hashes += _GlobalHashData      # the hash data from the file
      policyElement.Hashes.append(hashdata)
      print("%s - policyElement.Hashes size=0x%x" % (func, len(policyElement.Hashes))) # DBGDBG

    # update elementSize
    elementSize += DEFINES.DIGEST_SIZE[hashAlgName] * self.NumbHashes

    # ElementSize = 16 +HashSize* (PDEF.PolListInfo[i].StmDefData[index].NumbHashes)
    # if HashAlg == SHA-1, HashSize=20
    policyElement.ElementSize = elementSize
    print("%s - done, size=0x%x" % (func, elementSize))

    policyElements.append(policyElement)
    return(elementSize)

  def printDef(self, f):
    """printDef - write the Stm Def to the specified human readable text file for printing"""

    #print("printStmDef - object: %s" % (stmDefData))     # DBGDBG

    print("***STM_DEF***",                                      file=f)
    print("InfoSize",             " = ", self.InfoSize,         file=f)
    print("Tag",                  " = ", self.Tag,              file=f)
    #print("IncludeInList",        " = ", self.IncludeInList,    file=f)
    print("HashAlg",              " = ", hex(self.HashAlg),     file=f)
    print("Control",              " = ", self.Control,          file=f)
    print("NumbHashes",           " = ", self.NumbHashes,       file=f)
    print("CurrentView",          " = ", self.CurrentView,      file=f)
    print("HashFiles",            " = ", self.HashFiles,        file=f)


class PCONF_INFO( object ):
  """PCONF_INFO definition"""

  def __init__(self):
    """__init__() PCONF_INFO constructor"""

    self.pcrSelect            = [1, 0, 0]                  # UINT8 - pcrSelect[0]=1 => pcr0-7
    #reserved                 = 0                          # UINT8
    self.pcrFile                   = "myPcrFile.pcr"

    self.PCR0_BIT             = 0x01
    self.PCR1_BIT             = 0x02
    self.PCR2_BIT             = 0x04
    self.PCR3_BIT             = 0x08
    self.PCR4_BIT             = 0x10
    self.PCR5_BIT             = 0x20
    self.PCR6_BIT             = 0x40
    self.PCR7_BIT             = 0x80

class PCONF_DEF( object ):
  """PCONF_DEF class"""

  def __init__(self, hashAlg):
    """__init__() PCONF_DEF constructor"""
    #print("constructing a PCONF_DEF")

    hashAlgName = ""
    try:
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if hashAlg == val).next()
    except StopIteration:
      print("PCONF_DEF::__init__ - invalid hashAlg=%d" % (hashAlg))

    self.Name                = "PCONF-"+hashAlgName        # String - Name used for GUI identification
    self.InfoSize            = 24                          # UINT32 - number of bytes in this struct
    self.Tag                 = "PCON"                      # UINT32 - confirms that this is a PCONF policy defintion element
    self.IncludeInList       = False                       # BOOLEAN - indicates this LIST includes a PCONF element
    # Reserved5[6]           = 0                           # UINT8
    self.HashAlg             = hashAlg                     # UINT16 - TPM_ALG_SHAXXXX ...
    self.Control             = 0                           # UINT32 - USER: Bit0: Ignore PS PCONF elements
    self.NumbHashes          = 0                           # UINT16 - Tracks number of valid entries in PconfFiles[]
    self.CurrentView         = 0                           # UINT16 - User state: last PconfFiles[] selected
    #PCONF_INFO PcrInfoSrc[]                               # array of PCR selection and PCR dump file names
    self.PcrInfoSrc = []

  # Build the PCONF element and return its size
  # thisPdefList - is the list's source data from pdef.PolListInfo[list]
  # policyElement - is the destination LCP POLICY element
  # listCnt  - is the current list
  #
  def build(self, thisPdefList, policyElements, cwd):
    """buildPconfElement - build the PCONF element"""

    func = "buildPconfElement"
    print("%s" %(func)) # DBGDBG
    # build the element's data
    elementSize = PCONF_ELEMENT2_HDR_SIZE                   # size of all fields except PcrInfo's
    #pconfDefData = thisPdefList.PconfDefData[index]
    policyElement = LCP_PCONF_ELEMENT2()
    policyElement.PolEltControl = self.Control
    policyElement.HashAlg       = self.HashAlg
    policyElement.NumPCRInfos   = self.NumbHashes   # really number TPMS_QUOTE_INFO's

    # Build each policyElement.PCRInfos' TPMS_QUOTE_INFO
    # for each pcr
    i = 0                                                   # number of PcrInfoSrc's in LCP_PCONF_ELEMENT
    for pdefPcrInfo in self.PcrInfoSrc:
      thisTpmsQuoteInfo = TPMS_QUOTE_INFO()
      policyElement.PCRInfos.append(thisTpmsQuoteInfo)
      thisTpmsQuoteInfo = policyElement.PCRInfos[i]
      thisTpmlPcrSelection = thisTpmsQuoteInfo.pcrSelect    # TPML_PCR_SELECTION()
      thisTpm2bDigest      = thisTpmsQuoteInfo.pcrDigest    # TPM2B_DIGEST()
      # Set pcrSelections (which is type: TPMS_PCR_SELECTION) in a TPML_PCR_SELECTION to indicate the selected PCRs.
      # Hash is the same algorithm ID as HashAlg in LCP_PCONF_ELEMENT2 since TPML_PCR_SELECTION's count=1
      # Size of Select is always 3
      #   (because the TPM has 24 PCRs and thus requires 24-bits to indicate which PCR)
      # However the 2nd and 3rd bytes of pcrSelect[] are always 0x00 because the policy
      #   only selects between the first 8 PCRs.
      # pcrSelect[0] is a bit mask formed from thisPdefList.PconfDefData[index].PcrInfoSrc[i].pcrSelect[0]
      #   which is an 8 element list corresponding to each selected pcr from 0-7
      #
      thisTpmlPcrSelection.count = 1      # So 1 TPMS_PCR_SELECTION and HashAlg per PCONF, see Spec A.2.2.4 p31
      thisTpmsPcrSelection = thisTpmlPcrSelection.pcrSelections
      thisTpmsPcrSelection.hash = policyElement.HashAlg
      thisTpmsPcrSelection.sizeOfSelect = 0x0003

      pcr0to7SelectionBitMask = 0
      numSelectedPcrs = 0
      thisBit = 0x80
      pcr0to7SelectionList = pdefPcrInfo.pcrSelect[0]
      for eachBit in pcr0to7SelectionList:
        if(eachBit == 1):
            pcr0to7SelectionBitMask |= thisBit
            numSelectedPcrs += 1
        thisBit >>= 1

      thisTpmsPcrSelection.pcrSelect[0] =  pcr0to7SelectionBitMask
      thisTpmsPcrSelection.pcrSelect[1] =  0
      thisTpmsPcrSelection.pcrSelect[2] =  0
      print("%s - PCRInfos[%d], pcrSelect[0]=0x%x, numSelectedPcrs=%d" %
            (func, i, thisTpmsPcrSelection.pcrSelect[0], numSelectedPcrs)) # DBGDBG
      i += 1

      # determine if pcrFile is a PCRD or PCR2 formatted file
      file = pdefPcrInfo.pcrFile
      file = os.path.join(cwd, file)
      result = utilities.verifyPcrFile(file, policyElement.HashAlg)
      fileType = result[1]
      if(result[0] == False):
        print("%s - verifyPcrFile says PCR file %s is invalid!!!" % (func, file))  # Should NEVER get here
        return 0

      # calculate thisTpm2bDigest.size and build TPM2B_DIGEST.buffer from the PCR data
      # in this pdef list's selected pcrFiles
      hashAlg = self.HashAlg

      try:
        # reverse lookup of the hash algorithm name(key) for the given HashAlg value
        hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == hashAlg)).next()
      except StopIteration:
        print ("HashAlg=%d is not supported, aborting build" % (hashAlg))
        return 0

      hashSize = DEFINES.DIGEST_SIZE[hashAlgName]

      # verify that the pcfFile contains numHashes PCR values. These checks weren't done by verifyPcrFile()
      result = utilities.verifyPcrInfoNumHashes(file, fileType, hashAlg, pcr0to7SelectionBitMask)
      if(result == False):
        print("%s - verifyPcrInfoNumHashes says PCR file %s is invalid!!!" % (func, file))
        return 0

      elementSize += PCR_INFO2_HDR_SIZE  #  includes thisTpm2bDigest.size  but not buffer
      if(pcr0to7SelectionBitMask == 0):                               # no pcr's selected
        thisTpm2bDigest.size = 0                                      # buffer is empty

      else:
        #thisTpm2bDigest.size = numSelectedPcrs * hashSize             # sizeof(buffer)
        thisTpm2bDigest.size = hashSize                                # There's only one composite hash

        # for each selected PCR from 0 to 7
        if(pcr0to7SelectionBitMask != 0):
          #  Build tpmPcrSelection.digestAtRelease from the PCR data in this pdef list's selected pcrFiles
          hashdata = utilities.hashPcrInfoFromFile(os.path.join(cwd, file), pcr0to7SelectionBitMask, numSelectedPcrs)
          if(len(hashdata) != hashSize):
            print("%s - Invalid PCR file %s, aborting build" % (func, file))
            return 0

          tempbuf = array('B')
          tempbuf.fromstring(hashdata)
          #thisTpm2bDigest.buffer += _GlobalPcrHash
          thisTpm2bDigest.buffer += tempbuf
          elementSize += thisTpm2bDigest.size
        else:
          print ("%s - Nothing to hash: pcr0to7SelectionBitMask=0" %(func))  # DBGDBG
          print ("Note: No PCR was selected. If desired, select PCR's and click Update")

      #print("buildPconfElement - next pcrInfo i=%d" % (i)) # DBGDBG
      # end of pcrInfo for loop
    # update this element's size
    policyElement.ElementSize = elementSize
    print("%s - done, elementSize=0x%x" % (func, elementSize))

    policyElements.append(policyElement)
    return(elementSize)

  def printDef(self, f):
    """printDef - write the PCONF_DEF to the specified file"""

    #print("printPconfDef - object: %s" % (pconfDefData))     # DBGDBG

    print("***PCONF_DEF***",                                    file=f)
    print("InfoSize",             " = ", self.InfoSize,         file=f)
    print("Tag",                  " = ", self.Tag,              file=f)
    #print("IncludeInList",        " = ", self.IncludeInList,    file=f)
    print("HashAlg",              " = ", hex(self.HashAlg),     file=f)
    print("Control",              " = ", self.Control,          file=f)
    print("NumbHashes",           " = ", self.NumbHashes,       file=f)
    print("CurrentView",          " = ", self.CurrentView,      file=f)

    i = 0
    for eachPconfInfo in self.PcrInfoSrc:
      print("\n", file=f)         # for readability
      self.printPconfInfo(eachPconfInfo, i, f)
      i += 1

  def printPconfInfo(self, pconfInfo, index, f):
    """printPconfInfo - write the PCONF_INFO to the specified human readable text file for printing"""

    #print("printPconfInfo %d object: %s" % (index, pconfInfo))     # DBGDBG

    print("***PCONF_INFO", index, "***",                              file=f)
    print("InfoSize",             " = ", pconfInfo.pcrSelect,         file=f)
    print("Tag",                  " = ", pconfInfo.pcrFile,           file=f)


class PCONFLEGACY_DEF( object ):
  """PCONFLEGACY_DEF class"""

  def __init__(self):
    """__init__() PCONFLEGACY_DEF constructor"""
    #print("constructing a PCONFLEGACY_DEF")

    self.Name                = DEFINES.ELEMENT_NAME_PCONF_LEGACY # String - Name used for GUI identification
    self.InfoSize            = 24                          # UINT32 - number of bytes in this struct
    self.Tag                 = "PCON"                      # UINT32 - confirms that this is a PCONF policy defintion element
    self.IncludeInList       = False                       # BOOLEAN - indicates this LIST includes a PCONF element
    # Reserved5[6]           = 0                           # UINT8
    self.HashAlg             = 0                           # UINT8 - 0 = SHA1, others reserved for future use ...
    self.Control             = 0                           # UINT32 - USER: Bit0: Ignore PS PCONF elements
    self.NumbHashes          = 0                           # UINT16 - Tracks number of valid entries in PconfFiles[]
    self.CurrentView         = 0                           # UINT16 - User state: last PconfFiles[] selected
    #PCONF_INFO PcrInfoSrc[]                               # array of PCR selection and PCR dump file names
    self.PcrInfoSrc = []


  # Build the PCONF element and return its size
  # thisPdefList - is the list's source data from pdef.PolListInfo[list]
  # policyElement - is the destination LCP POLICY element
  #
  def build(self, thisPdefList, policyElements, cwd):
    """buildPconfLegacyElement - build the PCONF Legacy element"""

    func = "buildPconfLegacyElement"
    print("%s" %(func)) # DBGDBG
    elementSize = 14     # size of all fields except PcrInfo's

    # build the element's data
    #pconfDefData = thisPdefList.PconfLegacyDefData
    policyElement = LCP_PCONF_ELEMENT()
    policyElement.PolEltControl = self.Control
    policyElement.NumPCRInfos = self.NumbHashes

    # Build each policyElement.PCRInfos[NumPCRInfos]
    # for each hash to NumbHashes-1 i.e. to NumPCRInfos-1
    #
    i = 0     # number of PcrInfoSrc's in LCP_PCONF_ELEMENT
    for pdefPcrInfo in self.PcrInfoSrc:
        thisTpmPcrInfoShort = TPM_PCR_INFO_SHORT()
        policyElement.PCRInfos.append(thisTpmPcrInfoShort)
        thisTpmPcrInfoShort = policyElement.PCRInfos[i]
        thisTpmPcrInfoShort.localityAtRelease = 0x1f        # any locality

        # Set pcrSelection (which is type: TPM_PCR_SELECTION) to indicate the selected PCRs.
        # Size of Select is always 3
        #   (because the TPM has 24 PCRs and thus requires 24-bits to indicate which PCR)
        # However the 2nd and 3rd bytes of pcrSelect[] are always 0x00 because the policy
        #   only selects between the first 8 PCRs.
        # pcrSelect[0] is a bit mask formed from thisPdefList.PconfLegacyDefData[index].PcrInfoSrc[i].pcrSelect[0]
        #   which is an 8 element list corresponding to each selected pcr from 0-7
        #
        thisTpmPcrSelection = thisTpmPcrInfoShort.pcrSelection
        thisTpmPcrSelection.sizeOfSelect = 0x0003

        pcr0to7SelectionBitMask = 0
        numSelectedPcrs = 0
        thisBit = 0x80
        pcr0to7SelectionList = pdefPcrInfo.pcrSelect[0]
        for eachBit in pcr0to7SelectionList:
          if(eachBit == 1):
              pcr0to7SelectionBitMask |= thisBit
              numSelectedPcrs += 1

          thisBit >>= 1

        thisTpmPcrSelection.pcrSelect[0] =  pcr0to7SelectionBitMask
        thisTpmPcrSelection.pcrSelect[1] =  0
        thisTpmPcrSelection.pcrSelect[2] =  0

        print("%s - PCRInfos[%d], pcrSelect[0]=0x%x" % (func, i, thisTpmPcrSelection.pcrSelect[0])) # DBGDBG
        i += 1

        #  Build tpmPcrSelection.digestAtRelease from the PCR data in this pdef list's selected pcrFiles
        file = pdefPcrInfo.pcrFile
        hashdata = utilities.hashPcrInfoFromFile(os.path.join(cwd, file), pcr0to7SelectionBitMask, numSelectedPcrs)
        if(len(hashdata) != DEFINES.DIGEST_SIZE['SHA1']):
            print("%s - Invalid PCR file %s, aborting build" % (func, file))
            return 0

        thisTpmPcrInfoShort.digestAtRelease = hashdata      # hash of the selected PCR data from the file
        #print("buildPconfElement - digestAtRelease=%s, _GlobalPcrHash=%s" % ( thisTpmPcrInfoShort.digestAtRelease, _GlobalPcrHash))

    if(self.HashAlg == DEFINES.TPM_ALG_HASH['SHA1_LEGACY']):
      # ElementSize = 14 +PcrInfoSize* (PDEF.PolListInfo[i].PconfDefData[index].NumbHashes)
      # if HashAlg == SHA-1, PcrInfoSize=26
      elementSize += SHA1_PCR_INFO_SIZE * self.NumbHashes
    else:
      print ("HashAlg=%d is not supported, aborting build" % (thisPdefList.HashAlg))
      return 0

    # update this element's size
    policyElement.ElementSize = elementSize
    print("%s - done, size=0x%x" % (func, elementSize))

    policyElements.append(policyElement)
    return(elementSize)

  def printDef(self, f):
    """printPconfDef - write the PCONF_DEF to the specified file"""

    #print("printPconfDef - object: %s" % (pconfDefData))     # DBGDBG

    print("***PCONFLegacy_DEF***",                              file=f)
    print("InfoSize",             " = ", self.InfoSize,         file=f)
    print("Tag",                  " = ", self.Tag,              file=f)
    print("IncludeInList",        " = ", self.IncludeInList,    file=f)
    print("HashAlg",              " = ", hex(self.HashAlg),     file=f)
    print("Control",              " = ", self.Control,          file=f)
    print("NumbHashes",           " = ", self.NumbHashes,       file=f)
    print("CurrentView",          " = ", self.CurrentView,      file=f)

    i = 0
    for eachPconfInfo in self.PcrInfoSrc:
      print("\n", file=f)         # for readability
      self.printPconfInfo(eachPconfInfo, i, f)
      i += 1

  def printPconfInfo(self, pconfInfo, index, f):
    """printPconfInfo - write the PCONF_INFO to the specified human readable text file for printing"""

    #print("printPconfInfo %d object: %s" % (index, pconfInfo))     # DBGDBG

    print("***PCONF_INFO", index, "***",                              file=f)
    print("InfoSize",             " = ", pconfInfo.pcrSelect,         file=f)
    print("Tag",                  " = ", pconfInfo.pcrFile,           file=f)


class SBIOS_DEF( object ):
  """SBIOS_DEF class"""

  def __init__(self, hashAlg):
    """__init__() SBIOS_DEF constructor"""
    #print("constructing a SBIOS_DEF")

    hashAlgName = ""
    try:
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if hashAlg == val).next()
    except StopIteration:
      print("SBIOS_DEF::__init__ - invalid hashAlg=%d" % (hashAlg))

    self.Name                = "SBIOS-"+hashAlgName        # String - Name used for GUI identification
    self.InfoSize            = 56                          # UINT32 - number of bytes in this struct
    self.Tag                 = "SBIO"                      # UINT32 - confirms that this is a SBIOS policy defintion element
    self.IncludeInList       = False                       # BOOLEAN - indicates this LIST includes a SBIOS element
    # Reserved5[6]           = 0                           # UINT8
    self.HashAlg             = hashAlg                     # UINT16 - TPM_ALG_SHAXXXXX
    self.FallBackHashFile    = ""                          # FILEName - user - filename containing the fallback hash
    self.Control             = 0                           # UINT32 - USER: There are no defined controls at this time
    self.NumbHashes          = 0                           # UINT16 - Tracks number of valid entries in SbiosFiles[]
    self.CurrentView         = 0                           # UINT16 - User state: last PconfFiles[] selected
    self.SbiosFiles          = []                          # variable size array containing filenames of hashes

  # Build the SBIOS element and return its size
  # thisPdefList - is the list's source data from pdef.PolListInfo[list]
  # policyElement - is the destination LCP POLICY element
  # Return the elementSize built, or 0 if an error occurs
  #
  def build(self, thisPdefList, policyElements, cwd):
    """buildSbiosElement - build the SBIOS element"""

    func = "buildSbiosElement"
    print("%s"%(func)) # DBGDBG
    elementSize = 20

    # build the element's data
    #sbiosDefData = thisPdefList.SbiosDefData[index]
    policyElement = LCP_SBIOS_ELEMENT2()
    policyElement.PolEltControl = self.Control
    policyElement.HashAlg   = self.HashAlg
    invalidMsg = "Invalid fallbackhash file, aborting build"
    noneMsg    = "No fallbackhash file was specified, aborting build"

    hashAlg = self.HashAlg
    try:
      # reverse lookup of the hash algorithm name(key) for the given HashAlg value
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == hashAlg)).next()
    except StopIteration:
      print ("HashAlg=%d is not supported, aborting build" % (hashAlg))
      return 0

    # open the FallbackHash file and get the hash data
    file = self.FallBackHashFile
    if(file != ""):
      hashdata = utilities.getHashFromFile(os.path.join(cwd, file), self.HashAlg)
      if(len(hashdata) != DEFINES.DIGEST_SIZE[hashAlgName]):
        print("%s" % (invalidMsg))    # also show in output window
        return 0
    else:
      print("%s" % (noneMsg))         # also show in output window
      return 0

    # if there was no fallback hash file specified, then hash 0's [the init value of _GlobalHashData
    # otherwise use the value from that file
    #policyElement.FallbackHash = _GlobalHashData      # the hash data from the file
    policyElement.FallbackHash = hashdata
    policyElement.NumHashes    = self.NumbHashes

    # print("buildSbiosElement: type=%d, Alg=%d, Fallback=%s, GlobalHash=%s, NumbHashes=%d" %
    #        (policyElement.ElementType, policyElement.HashAlg ,
    #         policyElement.FallbackHash, _GlobalHashData, policyElement.NumHashes))     # DBGDBG

    # Build the hashes from each SbiosFiles[]
    for file in self.SbiosFiles:
      hashdata = utilities.getHashFromFile(os.path.join(cwd, file), self.HashAlg)
      if(len(hashdata) != DEFINES.DIGEST_SIZE[hashAlgName]):
        #self.StatusBar.SetStatusText("Invalid hash file %s, aborting build" % (file))
        return 0

      #policyElement.Hashes += _GlobalHashData      # the hash data from the file
      policyElement.Hashes.append(hashdata)
      print("%s - policyElement.Hashes size=0x%x" % (func,len(policyElement.Hashes))) # DBGDBG

    # update elementSize
    elementSize += DEFINES.DIGEST_SIZE[hashAlgName] * (self.NumbHashes+1) # +1 for fallback hash

    # ElementSize = 20 +HashSize* (1+PDEF.PolListInfo[i].SbiosDefData[index].NumbHashes)
    # if HashAlg == SHA-1, HashSize=20
    policyElement.ElementSize = elementSize
    print("%s - done, size=0x%x" % (func, elementSize))

    policyElements.append(policyElement)
    return(elementSize)

  def printDef(self, f):
    """printDef - write the Sbios Def to the specified human readable text file for printing"""

    #print("printSbiosDef - object: %s" % (sbiosDefData))     # DBGDBG

    print("***SBIOS_DEF***",                                    file=f)
    print("InfoSize",             " = ", self.InfoSize,         file=f)
    print("Tag",                  " = ", self.Tag,              file=f)
    #print("IncludeInList",        " = ", self.IncludeInList,    file=f)
    print("HashAlg",              " = ", hex(self.HashAlg),     file=f)
    print("FallBackHashFile",     " = ", self.FallBackHashFile, file=f)
    print("Control",              " = ", self.Control,          file=f)
    print("NumbHashes",           " = ", self.NumbHashes,       file=f)
    print("CurrentView",          " = ", self.CurrentView,      file=f)
    print("SbiosFiles",           " = ", self.SbiosFiles,       file=f)


class SBIOSLEGACY_DEF( object ):
  """SBIOS_DEF class"""

  def __init__(self):
    """__init__() SBIOSLEGACY_DEF constructor"""
    #print("constructing a SBIOSLEGACY_DEF")

    self.Name                = DEFINES.ELEMENT_NAME_SBIOS_LEGACY  # String - Name used for GUI identification
    self.InfoSize            = 56                          # UINT32 - number of bytes in this struct
    self.Tag                 = "SBIO"                      # UINT32 - confirms that this is a SBIOS policy defintion element
    self.IncludeInList       = False                       # BOOLEAN - indicates this LIST includes a SBIOS element
    # Reserved5[6]           = 0                           # UINT8
    self.HashAlg             = 0                           # UINT8 - 0 = SHA1, others reserved for future use ...
    self.FallBackHashFile    = ""                          # FILEName - user - filename containing the fallback hash
    self.Control             = 0                           # UINT32 - USER: There are no defined controls at this time
    self.NumbHashes          = 0                           # UINT16 - Tracks number of valid entries in SbiosFiles[]
    self.CurrentView         = 0                           # UINT16 - User state: last PconfFiles[] selected
    self.SbiosFiles          = []                          # variable size array containing filenames of hashes

  # Build the SBIOS element and return its size
  # thisPdefList - is the list's source data from pdef.PolListInfo[list]
  # policyElement - is the destination LCP POLICY element
  # Return the elementSize built, or 0 if an error occurs
  #
  def build(self, thisPdefList, policyElements, cwd):
    """buildSbiosElement - build the SBIOS element"""

    func = "buildSbiosLegacyElement"
    print("%s" %(func))
    elementSize = 20

    # build the element's data
    #sbiosDefData = thisPdefList.SbiosLegacyDefData[DEFINES.DEFDATA_INDEX_SHA1]
    policyElement = LCP_SBIOS_ELEMENT()
    policyElement.PolEltControl = self.Control
    policyElement.HashAlg   = self.HashAlg
    invalidMsg = "Invalid fallbackhash file, aborting build"
    noneMsg    = "No fallbackhash file was specified, aborting build"

    # open the FallbackHash file and get the hash data
    file = self.FallBackHashFile
    if(file != ""):
      hashdata = utilities.getHashFromFile(os.path.join(cwd, file), DEFINES.TPM_ALG_HASH['SHA1'])
      if(len(hashdata) != DEFINES.DIGEST_SIZE['SHA1']):
        print("%s" % (invalidMsg))    # also show in output window
        return 0
    else:
      print("%s" % (noneMsg))         # also show in output window
      return 0

    # if there was no fallback hash file specified, then hash 0's [the init value of _GlobalHashData
    # otherwise use the value from that file
    policyElement.FallbackHash = hashdata      # the hash data from the file
    #policyElement.Hashes.append(hashdata)

    # print("buildSbiosElement: type=%d, Alg=%d, Fallback=%s, GlobalHash=%s, NumbHashes=%d" %
    #        (policyElement.ElementType, policyElement.HashAlg ,
    #         policyElement.FallbackHash, _GlobalHashData, policyElement.NumHashes))     # DBGDBG

    policyElement.NumHashes    = self.NumbHashes
    # Build the hashes from each SbiosFiles[]
    for file in self.SbiosFiles:
      hashdata = utilities.getHashFromFile(os.path.join(cwd, file), DEFINES.TPM_ALG_HASH['SHA1'])
      if(len(hashdata) != DEFINES.DIGEST_SIZE['SHA1']):
        print ("Invalid hash file %s, aborting build" % (file))
        return 0

      policyElement.Hashes.append(hashdata)      # the hash data from the file
      print("%s - policyElement.Hashes size=0x%x" % (func, len(policyElement.Hashes))) # DBGDBG

    if(self.HashAlg == DEFINES.TPM_ALG_HASH['SHA1_LEGACY']):
      # update elementSize
      elementSize += DEFINES.DIGEST_SIZE['SHA1'] * (self.NumbHashes+1)   # +1 for fallback hash
    else:
      print("%s - HashAlg=%d is not supported, aborting build" % (func, thisPdefList.HashAlg))
      return 0

    # ElementSize = 20 +HashSize* (1+PDEF.PolListInfo[i].SbiosDefData.NumbHashes)
    # if HashAlg == SHA-1, HashSize=20
    policyElement.ElementSize = elementSize
    print("%s - done, size=0x%x" % (func, elementSize))

    policyElements.append(policyElement)
    return(elementSize)

  def printDef(self, f):
    """printSbiosDef - write the Sbios Def to the specified human readable text file for printing"""

    #print("printSbiosDef - object: %s" % (sbiosDefData))     # DBGDBG

    print("***SBIOSLEGACY_DEF***",                              file=f)
    print("InfoSize",             " = ", self.InfoSize,         file=f)
    print("Tag",                  " = ", self.Tag,              file=f)
    #print("IncludeInList",        " = ", self.IncludeInList,    file=f)
    print("HashAlg",              " = ", hex(self.HashAlg),     file=f)
    print("FallBackHashFile",     " = ", self.FallBackHashFile, file=f)
    print("Control",              " = ", self.Control,          file=f)
    print("NumbHashes",           " = ", self.NumbHashes,       file=f)
    print("CurrentView",          " = ", self.CurrentView,      file=f)
    print("SbiosFiles",           " = ", self.SbiosFiles,       file=f)



#!/usr/bin/python
#  Copyright (c) 2013, Intel Corporation. All rights reserved.

# TXT Policy Generator Tool
# support for Building LCP_POLICY and LCP_POLICY_DATA Files

# using print() built in function, disable print statement
from __future__ import print_function

from struct import *
#import hashlib
import binascii

from defines import DEFINES
from pdef import PDEF

from util import UTILS
utilities = UTILS()

from LcpPolicy import *
LcpPolicy = LCP_POLICY2()
LcpPolicyData = LCP_POLICY_DATA2()

ListMeasurements = ListMeasurements()         # DIGEST ListMeasurements[PDEF.NumLists]

try:
  import os
  import sys
except ImportError:
  raise ImportError, "import OS failed"

import array
import M2Crypto

from asn1spec import *


class Build( object ):
  """build LCP_POLICY2 and LCP_POLICY_DATA2 Files"""

  filename = "MyPlatformXXXXXXXXXXXXX"
  dirname  = os.getcwd()

  def __init__( self ):
    """Build - constructor"""
    pass

  # Build the .pol file
  # returns True if all went ok
  # returns False on error
  def buildRawLcpPolicyFile(self, pdef, statusBar):
    """buildRawLcpPolicyFile - build the raw format LCP_POLICY2 file (.pol)"""

    self.StatusBar = statusBar

    # if policy version is 3.0, clear bit positions for ignore PS elements
    policyversion = str(pdef.PolVersionMajor)+'.'+str(pdef.PolVersionMinor)
    if policyversion == '3.0':
      mask = DEFINES.PolicyControlIgnorePsMleBitMask | DEFINES.PolicyControlIgnorePsPconfBitMask | DEFINES.PolicyControlIgnorePsStmBitMask
      pdef.PolicyControl &= ~mask

    # pack the LCP_POLICY2 struct [without the hash] into a binary string ready for the LCP_POLICY2 file
    packedLcpPolicy = pack(LcpPolicy.LcpPolicyFormatStringNoHash,
                     pdef.PolVersionMinor, pdef.PolVersionMajor,
                     pdef.HashAlg, pdef.PolicyType, pdef.SinitMinVersion,
                     pdef.DataRevocationCounters[0], pdef.DataRevocationCounters[1],
                     pdef.DataRevocationCounters[2], pdef.DataRevocationCounters[3],
                     pdef.DataRevocationCounters[4], pdef.DataRevocationCounters[5],
                     pdef.DataRevocationCounters[6], pdef.DataRevocationCounters[7],
                     pdef.PolicyControl,
                     pdef.MaxSinitMinVersion, pdef.MaxBiosMinVersion,
                     pdef.LcpHashAlgMask, pdef.LcpSignAlgMask, pdef.AuxHashAlgMask, 0)

    # reuse the base name [with .pol extension] and don't prompt for a file name again
    basefilename, ext = self.filename.rsplit('.', 1)
    #polfilename = utilities.formFileName(basefilename, "pol")
    polfilename = ".".join([basefilename, 'pol'])
    print("buildRawLcpPolicyFile:  creating Text LCP_POLICY file %s" % (polfilename))  # DBGDBG

    # write the raw LCP_POLICY file
    #   1st write the struct, then write the hash from pdef.PolicyHash
    try:
      f = open(os.path.join(self.dirname, polfilename), 'wb')
      print(packedLcpPolicy,     end='', file=f )             # suppress EOL (0x0a)

      try:
        # reverse lookup of the hash algorithm name(key) for the given HashAlg value
        hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == pdef.HashAlg)).next()

        # If PolicyType==ANY, haven't updated pdef.PolicyHash (since there were no lists to process)
        # In that case, write a hash of all 0's
        if(pdef.PolicyType == DEFINES.ANY):
          # Pack hash value which is 0 for PolicyType of ANY
          b = bytes()
          pdef.PolicyHash = b.join(pack('B', 0) for i in range(DEFINES.DIGEST_SIZE[hashAlgName]))
        else:
          # Write packed hash value to .pol file.
          print(pdef.PolicyHash, end='', file=f )

      except StopIteration:
        # pdef.HashAlg value is not found in DEFINES.TPM_ALG_HASH list
        self.StatusBar.SetStatusText("Aborting build, Hash Algorithm %d is not supported." % (pdef.HashAlg))
        print("buildRawLcpPolicyFile: ANY - aborting - Hash Algorithm %d is not supported." % (pdef.HashAlg))  # DBGDBG
        f.close()
        return False

      # If PolicyType==ANY, haven't updated pdef.PolicyHash (since there were no lists to process)
      # In that case, write a hash of all 0's
      #if(pdef.PolicyType == DEFINES.ANY):
      #  # Pack hash value which is 0 for PolicyType of ANY
      #  if (hashAlgName != None):
      #    b = bytes()
      #    pdef.PolicyHash = b.join(pack('B', 0) for i in range(DEFINES.DIGEST_SIZE[hashAlgName]))
      #  else:
      #    self.StatusBar.SetStatusText("Aborting build, Hash Algorithm %d is not supported." % (pdef.HashAlg))
      #    print("buildRawLcpPolicyFile: ANY - aborting - Hash Algorithm %d is not supported." % (pdef.HashAlg))  # DBGDBG
      #    f.close()
      #    return False
          
        #if(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SHA1']):
        #  pdef.PolicyHashSha1   = pack("20B", 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
        #elif(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SHA256']):
        #  pdef.PolicyHashSha256 = pack("32B", 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
        #elif(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SHA384']):
        #  pdef.PolicyHashSha384 = pack("40B", 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
        #elif(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SM3']):
        #  pdef.PolicyHashSm3    = pack("32B", 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
        #else:
        #  self.StatusBar.SetStatusText("Aborting build, Hash Algorithm %d is not supported." % (pdef.HashAlg))
        #  print("buildRawLcpPolicyFile: ANY - aborting - Hash Algorithm %d is not supported." % (pdef.HashAlg))  # DBGDBG
        #  f.close()
        #  return False

      
      # Pack hash value which is 0 for PolicyType of ANY
      #if (hashAlgName != None):
      #  # Write packed hash value to .pol file.
      #  print(pdef.PolicyHash, end='', file=f )
      #else:     # don't hit the above else case if policyType=LIST
      #    self.StatusBar.SetStatusText("Aborting build, Hash Algorithm %d is not supported." % (pdef.HashAlg))
      #    print("buildRawLcpPolicyFile: LIST - aborting - Hash Algorithm %d is not supported." % (pdef.HashAlg))  # DBGDBG
      #    f.close()
      #    return False
      #if(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SHA1']):
      #  print(pdef.PolicyHashSha1, end='', file=f )
      #elif(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SHA256']):
      #  print(pdef.PolicyHashSha256, end='', file=f )
      #else:     # don't hit the above else case if policyType=LIST
      #    self.StatusBar.SetStatusText("Aborting build, Hash Algorithm %d is not supported." % (pdef.HashAlg))
      #    print("buildRawLcpPolicyFile: LIST - aborting - Hash Algorithm %d is not supported." % (pdef.HashAlg))  # DBGDBG
      #    f.close()
      #    return False

    except IOError:
      print ("IOError writing raw LCP_POLICY file %s" % (polfilename))
    finally:
      f.close()

    print("buildRawLcpPolicyFile:  Raw LCP_POLICY file %s built" % (polfilename))  # DBGDBG

    # originally, filename had no extention so strip it off
    #self.filename, ext = self.filename.rsplit('.', 1)
    return True

  def buildTxtLcpPolicyFile(self, pdef):
    """buildTxtLcpPolicyFile - build the text format LCP_POLICY2 file"""

    # reuse the base name [with .txt extention] and don't prompt for a file name again
    basefilename, ext = self.filename.rsplit('.', 1)
    txtfilename = utilities.formFileName(basefilename, "txt")
    #self.filename = utilities.formFileName(self.filename, "txt")
    print("buildTxtLcpPolicyFile:  creating Text LCP_POLICY2 file %s" % (txtfilename))  # DBGDBG

    try:
      f = open(os.path.join(self.dirname, txtfilename), 'w')

      # suppress separator with sep=''
      # use 'fmt' % (value) to prepend 0's to value

      print("#Policy Data", file=f)

      version = '%02d%02d' % (pdef.PolVersionMajor, pdef.PolVersionMinor)
      print("word=", version, "\t\t# Version", sep='', file=f)

      hashAlg = '%04d' % (pdef.HashAlg)
      print("word=", hashAlg, "\t\t# Hash Alg (4=Sha1[20], 11=Sha256[32]", sep='', file=f)

      policyType = '%02d' % (pdef.PolicyType)
      print("byte=", policyType, "\t\t\t# PolicyType (0 = LIST, 1 = ANY)", sep='', file=f)

      sinitMinVersion = '%02d' % (pdef.SinitMinVersion)
      print("byte=", sinitMinVersion, "\t\t\t# SinitMinVersion", sep='', file=f)

      i = 0
      while(i < pdef.MaxLists):
        dataRevCnt = '%04d' % (pdef.DataRevocationCounters[i])
        print("word=", dataRevCnt, "\t\t# DataRevocationCounters[", i, "]", sep='', file=f)
        i += 1

      policyControl = '0x%08x' % (pdef.PolicyControl)
      print("dword=", policyControl, "\t# PolicyControl (Bit 1=Allow NPW, 2=PCR17, 3=Force Owner 15=Aux Delete)", sep='', file=f)

      maxSinitMinVersion = '%02d' % (pdef.MaxSinitMinVersion)
      print("byte=", maxSinitMinVersion, "\t\t\t# MaxSinitMinVersion", sep='', file=f)
      maxBiosMinVersion = '%02d' % (pdef.MaxBiosMinVersion)
      print("byte=", maxBiosMinVersion, "\t\t\t# MaxBiosMinVersion", sep='', file=f)
      
      lcpHashAlgMask = '%04d' % (pdef.LcpHashAlgMask)
      print("word=", lcpHashAlgMask, "\t\t# LcpHashAlgMask", sep='', file=f)
      lcpSignAlgMask = '%04d' % (pdef.LcpSignAlgMask)
      print("dword=", lcpSignAlgMask, "\t\t# LcpSignAlgMask", sep='', file=f)
      auxHashAlgMask = '%04d' % (pdef.AuxHashAlgMask)
      print("word=", auxHashAlgMask, "\t\t# AuxHashAlgMask", sep='', file=f)

      reserved = '%02d' % (0)
      print("byte=", reserved, "\t\t\t# Reserved", sep='', file=f)
      
      #if(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SHA256']):
      #  print("byte=", pdef.PolicyHashSha256Hex, " # Hash", sep='', file=f)   # need to print the hex hash
      #elif(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SHA1']):
      #  print("byte=", pdef.PolicyHashSha1Hex,   " # Hash", sep='', file=f)   # need to print the hex hash
      print("byte=", binascii.b2a_hex(pdef.PolicyHash), " # Hash", sep='', file=f)   # need to print the hex hash
      
    except IOError:
      print ("IOError writing file %s" % (txtfilename))
    finally:
      f.close()

    print("buildTxtLcpPolicyFile:  Text LCP_POLICY file %s built" % (txtfilename))  # DBGDBG

    # originally, filename had no extention so strip it off
    #self.filename, ext = self.filename.rsplit('.', 1)

  # LCP2 does not generate an XML output file - so no buildXmlLcpPolicyFile()

  # If no error, return True, else False
  # Build can fail reading hash or pcr files, or if no public or private key file was selected
  #
  def buildLcpPolicyDataStruct(self, pdef, statusBar):
    """buildLcpPolicyDataStruct - create a LCP_POLICY_DATA struct including calculating the PolicyHash"""

    global LcpPolicyData

    self.StatusBar = statusBar
    #LcpPolicy = LCP_POLICY2()
    LcpPolicyData = LCP_POLICY_DATA2()
    # ListMeasurements - object to hold ListMeasurements[pdef.NumLists]
    # LcpPolicyData    - object for LCP_POLICY_DATA
    # build file header: FileSignature,0,0,0,NumLists
    LcpPolicyData.NumLists = pdef.NumLists

    print("buildLcpPolicyDataStruct: LCP_POLICY_DATA: FileSignature=%s, NumLists=%d" %
          (LcpPolicyData.FileSignature, LcpPolicyData.NumLists)) # DBGDBG

    # For each list
    #   If 1st list, build the list header: ListVersion, 0, SigAlgorithm, PolicyElementsSize=0
    #   build the elements
    #   if List is unsigned, hash the list
    #   else list is signed so build the signature block   ie:
    #   Done with this list
    # Done with all the lists
    policyElementsSize = 0            # cumulative size of all the elements in this list
    listCnt = 0
    listCntStr = str(listCnt)
    while(listCnt < pdef.NumLists):
      thisPdefList = pdef.PolListInfo[listCntStr]
      thisPolicyList = LcpPolicyData.PolicyLists[listCntStr]

      # build the LCP_POLICY_LIST.PolicyList[N] header from pdef.PolListInfo[N]
      thisPolicyList.VersionMajor = thisPdefList.ListVersionMajor
      thisPolicyList.VersionMinor = thisPdefList.ListVersionMinor
      thisPolicyList.SigAlgorithm = thisPdefList.SigAlgorithm
      thisPolicyList.PolicyElementsSize = 0            # TBD

      # if the list is signed, make sure that a public and private key file was specified
      # otherwise abort the build
      #if(thisPdefList.SigAlgorithm == DEFINES.LCP_POLSALG_RSA_PKCS_15):     # TODO: add external signature support
      #if(thisPdefList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['RSASSA']):
      if thisPdefList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['NULL']:
        # Not Signing..
        pass
      elif thisPdefList.SigAlgorithm in DEFINES.TPM_ALG_SIGN.values():
        if((thisPdefList.PubKeyFile == "") or (thisPdefList.PvtKeyFile == "") and not thisPdefList.PvtKeyFileIsSignature):
          self.StatusBar.SetStatusText("Build failed: does not have both a public and private key file selected")
          print("buildLcpPolicyDataStruct failed - see status bar") # DBGDBG
          return False
      else:
        # What about invalid signature algorithms.
        print("ERROR: undefined signature algorithm")
        pass

      #IncludeElement = utilities.getIncludeElement(thisPdefList)
      #i=0
      #flag = False
      #while(i < len(IncludeElement)):
      #  if(IncludeElement[i] == True):
      #    flag = True
      #  i += 1

      # if there are elements to build, then build them and update PolicyElementsSize
      #if(flag == True):
      if len(thisPdefList.ElementDefData):
        thisPolicyList.PolicyElementsSize = self.buildListElements(pdef, thisPdefList, thisPolicyList, listCnt)
        # check for errors building the elements, i.e. reading the hash or pcr files
        if(thisPolicyList.PolicyElementsSize == 0):
          print("buildLcpPolicyDataStruct failed: PolicyElementsSize=0")  # DBGDBG
          return False

      if not self.hashListOrBuildSignatureBlock(pdef, thisPdefList, thisPolicyList, listCnt):
        return False
      print("buildLcpPolicyDataStruct: LCP_POLICY_LIST: Version=%x.%x, SigAlgorithm=%d, PolicyElementsSize=%d" %
          (thisPolicyList.VersionMajor, thisPolicyList.VersionMinor,
           thisPolicyList.SigAlgorithm, thisPolicyList.PolicyElementsSize)) # DBGDBG

      thisPdefList.ListModified = False
      listCnt += 1
      listCntStr = str(listCnt)

    try:
      # reverse lookup of the hash algorithm name(key) for the given HashAlg value
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == pdef.HashAlg)).next()
    except StopIteration:
      print("Aborting build, Hash Algorithm %d is not supported" % (pdef.HashAlg))
      return False

    # Calculate list measurements - hash ListMeasurements[] using pdef.HashAlg

    hash = M2Crypto.EVP.MessageDigest(hashAlgName.lower())

    listCnt = 0
    while(listCnt < pdef.NumLists):
      #if(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SHA1']):
      #  thisHash = ListMeasurements.hashes[str(listCnt)]
      #elif(pdef.HashAlg == DEFINES.TPM_ALG_HASH['SHA256']):
      #  thisHash = ListMeasurements.hashes32[str(listCnt)]
      thisHash = ListMeasurements.hashes[str(listCnt)]
      
      hash.update(bytes(thisHash))
      #print("buildLcpPolicyDataStruct: ListMeasurement[%d] hash=%s" % (listCnt, thisHash)) # DBGDBG
      print("buildLcpPolicyDataStruct: ListMeasurement[%d] hashed" % (listCnt)) # DBGDBG
      listCnt += 1

    # save to pdef.PolicyHash
    pdef.PolicyHash = hash.digest()

    #print("buildLcpPolicyDataStruct - hash size=%d hexdigest=%s" % (hash.digest_size, hash.hexdigest()))   # DBGDBG
    pdef.Modified = False
    return True

  # Build the elements and return their size
  # If an error occurred, size = 0
  # Build can fail reading hash or pcr files
  #
  def buildListElements(self, pdef, thisPdefList, thisPolicyList, listCnt):
    """buildElements - build this lists elements"""
    func = 'buildListElements'

    print("%s: list %d - MLE, PCONF then SBIOS" % (func, listCnt))  # DBGDBG
    policyElementsSize = 0
    thisElementSize = 0

    # Sort element in this order [MLE, STM, SBIOS, PCONF] from [SHA512... SHA1, SHA1-LEGACY]
    sortedPdefList = []
    mlelist = [element for element in thisPdefList.ElementDefData if 'MLE' in element.Name]
    sortedPdefList += sorted(mlelist, key=lambda x: x.HashAlg, reverse=True)
    stmlist = [element for element in thisPdefList.ElementDefData if 'STM' in element.Name]
    sortedPdefList += sorted(stmlist, key=lambda x: x.HashAlg, reverse=True)
    sbioslist = [element for element in thisPdefList.ElementDefData if 'SBIOS' in element.Name]
    sortedPdefList += sorted(sbioslist, key=lambda x: x.HashAlg, reverse=True)
    pconflist = [element for element in thisPdefList.ElementDefData if 'PCONF' in element.Name]
    sortedPdefList += sorted(pconflist, key=lambda x: x.HashAlg, reverse=True)

    elementCnt = 0
    for element in sortedPdefList:
      print("building Element %s" %(element.Name))
      # call the element's build function to build the elements of this policy
      thisElementSize = element.build(thisPdefList, thisPolicyList.PolicyElements, pdef.WorkingDirectory)
      policyElementsSize += thisElementSize
      elementCnt += 1

    print("%s: list %d - elementCnt=%d, policyElementsSize=0x%x" % (func, listCnt, elementCnt, policyElementsSize)) # DBGDBG
    return(policyElementsSize)


  # Build the .dat file
  def buildLcpPolicyDataFile(self, pdef):
    """buildLcpPolicyDataFile - build the LCP_POLICY_DATA2 file (.dat)"""
    func = 'buildLcpPolicyDataFile'

    # reuse the base name [with .dat extention] and don't prompt for a file name again
    basefilename, ext = self.filename.rsplit('.', 1)
    datfilename = utilities.formFileName(basefilename, "dat")
    #self.filename = utilities.formFileName(self.filename, "dat")
    #lcpPolicyDataHeader = self.packLcpPolicyDataHeader()
    lcpPolicyDataHeader = LcpPolicyData.pack()

    # write the LCP_POLICY_DATA header to the .dat file: f
    # write the data to be signed to tmpFile: tmp
    #
    tmpFile = "tmpLcpPolicyList.tmp"
    try:
      f   = open(os.path.join(self.dirname, datfilename), 'wb')
      print(lcpPolicyDataHeader, end='', file=f )             # suppress EOL (0x0a)
    except IOError:
      self.StatusBar.SetStatusText("IOError writing raw LCP_POLICY_DATA file %s" % (datfilename))
      return

    # For each List in LcpPolicyData.PolicyLists[]
    print("buildLcpPolicyDataFile:  creating LCP_POLICY_DATA file %s" % (datfilename))  # DBGDBG
    listCnt = 0
    listCntStr = str(listCnt)
    while(listCnt < pdef.NumLists):
      try:
        tmp = open(os.path.join(self.dirname, tmpFile), 'wb')
      except IOError:
        self.StatusBar.SetStatusText("IOError opening tempFile %s" % (tmpFile))
        return
      print("buildLcpPolicyDataFile: processing list %d of %d" % (listCnt, pdef.NumLists)) # DBGDBG
      thisPolicyList = LcpPolicyData.PolicyLists[listCntStr]
      #lcpPolicyListHeader = self.packLcpPolicyListHeader(thisPolicyList)
      lcpPolicyListHeader = thisPolicyList.pack()

      # There can be 1 element of each type of each hash algorithm, See LCP_POLICY_LIST2::PolicyElements[]
      lcpPolicyListElement = [None, None, None, None, None, None, None, None]

      # pack this list's elements, if they exist
      i=0
      numElements = len(thisPolicyList.PolicyElements)
      while(i < numElements):
        #print("%s - checking if PolicyElement %d of %d exists" % (func, i, numElements)) # DBGDBG
        if(thisPolicyList.PolicyElements[i] != None):
          print("%s - packing PolicyElement %d of %d" % (func, i, numElements)) # DBGDBG
          #lcpPolicyListElement[i] = self.packLcpPolicyElement(pdef, str(listCnt), i, thisPolicyList.PolicyElements[i])
          lcpPolicyListElement[i] = thisPolicyList.PolicyElements[i].pack()
        else:
          break             # done after all existing elements are packed
        i += 1

      # write this LCP_POLICY_LIST to the .dat file
      #   1st write the LcpPolicyListHeader,
      #   then write the elements and optionally the signature (if list is signed)
      #
      try:
        print(lcpPolicyListHeader, end='', file=f )
        print(lcpPolicyListHeader, end='', file=tmp )
        # print the this list's elements, if they exist
        i=0
        while(i < numElements):
          #print("%s - checking if PolicyElement %d of %d was packed" % (func, i, numElements)) # DBGDBG
          if(lcpPolicyListElement[i] != None):
            print("%s - writing PolicyElement %d of %d to the file" % (func, i, numElements)) # DBGDBG
            print(lcpPolicyListElement[i], end='', file=f )
            print(lcpPolicyListElement[i], end='', file=tmp )
          i += 1

      except IOError:
        print ("IOError writing raw LCP_POLICY_DATA file %s" % (datfilename))
        f.close()
        return

      # if list is signed, now that we have formed all the binary data, it can be signed
      # and the LCP_SIGNATURE object can be written to the file
      #
      # The RSA signature is calculated over the entire LCP_POLICY_LIST struct,
      # including the Signature member, EXCEPT for the SigBlock
      #
      #if(thisPolicyList.SigAlgorithm == DEFINES.LCP_POLSALG_RSA_PKCS_15):
      if(thisPolicyList.SigAlgorithm != DEFINES.TPM_ALG_SIGN['NULL']):
        print("buildLcpPolicyDataFile: sign list %d" % (listCnt)) # DBGDBG
        self.signThisList(pdef, listCntStr, thisPolicyList, f, tmp, tmpFile)

      listCnt += 1
      listCntStr = str(listCnt)

    f.close()
    print("buildLcpPolicyDataFile:  LCP_POLICY_DATA file %s built" % (datfilename))  # DBGDBG


  def signThisList(self, pdef, listCntStr, thisPolicyList, f, tmp, tmpFile):
    """signThisList - form the signature of the specified list and write it to the specified files"""

    # get the private key from its file
    thisPdefList = pdef.PolListInfo[listCntStr]

    signAlgStr = ""
    try:
      signAlgStr = (key for key,val in DEFINES.TPM_ALG_SIGN.items() if val == thisPdefList.SigAlgorithm).next()
    except StopIteration:
      print ("Unsupported signature algorithm (%d)" %(thisPdefList.SigAlgorithm))
      return

    # pack the LCP_SIGNATURE2 object except for the PublicKeyValue and SigBlock and write it to the file
    # those 2 member's size depends on keySize so they are written separately
    lcpSignature = thisPolicyList.Signature.pack()
    print(lcpSignature, end='', file=f )
    print(lcpSignature, end='', file=tmp )

    # Append Public key to signature block.
    if 'RSA' in signAlgStr:
      print(thisPolicyList.Signature.PubkeyValue, end='', file=f)
      print(thisPolicyList.Signature.PubkeyValue, end='', file=tmp)
    else:
      # ECC
      print(thisPolicyList.Signature.Qx, end='', file=f)
      print(thisPolicyList.Signature.Qx, end='', file=tmp)
      print(thisPolicyList.Signature.Qy, end='', file=f)
      print(thisPolicyList.Signature.Qy, end='', file=tmp)
    tmp.close()               # done writing to tmp

    # read back the list data from the tmp file so it can be hashed and signed
    tmp = open(os.path.join(self.dirname, tmpFile), 'rb')
    tmp.seek (0, os.SEEK_END)
    listFileSize = tmp.tell ()
    tmp.seek (0, os.SEEK_SET)           # back to the beginning
    tmpListData = array.array ("B")     # Load file into data array
    try:
      tmpListData.fromfile (tmp, listFileSize)
    except:
      self.StatusBar.SetStatusText("Error reading list data from tmp file")
      print("signThisList:  ******Error reading list data from tmp file!!!!!!" )  # DBGDBG ******
      tmp.close()
      return

    tmp.close()

    print("signThisList:  hashing %d bytes from %s" % (listFileSize, tmpFile))  # DBGDBG
    # hash the list and sign the hash
    signatureBE = None
    hashAlgName = ""
    try:
      # reverse lookup of the hash algorithm name(key) for the given HashAlg value
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == thisPdefList.sigAlgorithmHash)).next()
    except StopIteration:
      self.StatusBar.SetStatusText("Aborting build, Hash Algorithm %d is not supported." % (thisPdefList.sigAlgorithmHash))
      print("signThisList: aborting - Hash Algorithm %d is not supported." % (thisPdefList.sigAlgorithmHash))  # DBGDBG
      return

    h = None      # set hash value to null first
    MsgDigest = M2Crypto.EVP.MessageDigest(hashAlgName.lower())
    MsgDigest.update(tmpListData)
    h = MsgDigest.digest()

    # Output hash for each list to a file for CA signing
    basefilename, ext = self.filename.rsplit('.', 1)
    hashfilename = basefilename + '_list' + listCntStr + '.hash'
    hashout = open(os.path.join(self.dirname, hashfilename), 'wb')
    hashout.write(h)
    hashout.close()
    # Output list raw binary output to a file for CA hashing and signing
    listbinfilename = basefilename + '_list' + listCntStr + '.bin'
    listbinout = open(os.path.join(self.dirname, listbinfilename), 'wb')
    listbinout.write(tmpListData)
    listbinout.close()


    # Write signature data.
    if thisPdefList.PvtKeyFileIsSignature:
      sig_str = ""
      if thisPdefList.PvtKeyFile != "":
        # Load signature using the file in private key
        fsig = open(os.path.join(self.dirname, thisPdefList.PvtKeyFile), 'rb')
        fsig.seek(0, 2)
        size = fsig.tell()
        fsig.seek(0, 0)
        sig_array = array.array('B')
        sig_array.fromfile(fsig, size)
        sig_str = sig_array.tostring()
        fsig.close()
      else:
        # if no file defined, then it's the first pass, before the list is signed by certificate authority(CA)
        pass

      if 'RSA' in signAlgStr:

        signatureBE = sig_str
        # byte reverse the signature, i.e. convert from big to little-endian
        signatureLE = signatureBE[::-1]
        thisPolicyList.Signature.SigBlock = signatureLE
        #print("signThisList:  LCP_POLICY_LIST %d sigLen=0x%x sig=\n%s" % (listCnt, len(signature), signature))  # DBGDBG

        # write the LCP_SIGNATURE.SigBlock object to the file
        print(signatureLE, end='', file=f )
        print("signThisList: Finished writing signature for list %d" % (int(listCntStr))) # DBGDBG
      else:
        # ECC signatures
        signature, substrate = der_decoder.decode(sig_str, asn1Spec=ECDSASignature())

        ri = signature.getComponentByName('r')
        si = signature.getComponentByName('s')

        roctet = univ.OctetString(hexValue=format(int(ri), '0x'))
        soctet = univ.OctetString(hexValue=format(int(si), '0x'))

        rBE = roctet.asNumbers()
        sBE = soctet.asNumbers()

        rLE = rBE[::-1]
        sLE = sBE[::-1]
        thisPolicyList.Signature.r = rLE
        thisPolicyList.Signature.s = sLE
        # write the LCP_SIGNATURE.SigBlock object to the file
        print(rLE, end='', file=f )
        print(sLE, end='', file=f )

    else:
      # Read Private Key from file
      # then sign the list
      #key = RSA.importKey(open(os.path.join(self.dirname, thisPdefList.PvtKeyFile)).read())
      key = None
      mb = None
      with open(os.path.join(self.dirname, thisPdefList.PvtKeyFile), 'rb') as kf:
        mb = M2Crypto.BIO.MemoryBuffer(kf.read())
      if mb:
        if 'RSA' in signAlgStr:
          key = M2Crypto.RSA.load_key_bio(mb)
        elif 'EC' in signAlgStr:
          # if ECC algorithm
          key = M2Crypto.EC.load_key_bio(mb)
      else:
        print("DEBUG: error load key")
      kf.close()

      #print("signThisList:  LCP_POLICY_LIST %s pvt key=\n%s" % (listCntStr, key))  # DBGDBG

      if 'RSA' in signAlgStr:
        #signer = PKCS1_v1_5.new(key)
        #signatureBE = signer.sign(h)
        signatureBE = key.sign(h, hashAlgName.lower())
        # byte reverse the signature, i.e. convert from big to little-endian
        signatureLE = signatureBE[::-1]
        thisPolicyList.Signature.SigBlock = signatureLE
        #print("signThisList:  LCP_POLICY_LIST %d sigLen=0x%x sig=\n%s" % (listCnt, len(signature), signature))  # DBGDBG

        # write the LCP_SIGNATURE.SigBlock object to the file
        print(signatureLE, end='', file=f )
      else:
        # ECC signatures
        r, s = key.sign_dsa(h)
        #print("DEBUG: ECC signature size r(%d) s(%d)" %(len(rBE), len(sBE)))

        length = int(thisPdefList.KeySize)/8
        rBE = r[-length:]
        sBE = s[-length:]
        rLE = rBE[::-1]
        sLE = sBE[::-1]
        thisPolicyList.Signature.r = rLE
        thisPolicyList.Signature.s = sLE
        # write the LCP_SIGNATURE.SigBlock object to the file
        print(rLE, end='', file=f )
        print(sLE, end='', file=f )

    print("signThisList: Finished writing signature for list %d" % (int(listCntStr))) # DBGDBG


  def packLcpSignatureHdr(self, lcpSignature):
    """packLcpSignature - pack this LCP_SIGNATURE object except its SigBlock member ad return the packed data"""

    # Note - The public key size in the pdef is the number of bits (1024, 2048, 3072)
    #        while LCP_SIGNATURE.PublicKeySize is the number if bytes, hence the /8 below
    return(pack(lcpSignature.LcpSignatureHdrFormatString,
                lcpSignature.RevocationCounter, lcpSignature.PubkeySize/8))

  def hashListOrBuildSignatureBlock(self, pdef, thisPdefList, thisPolicyList, listCnt)  :
    """hashListOrBuildSignatureBlock - if list unsigned then hash it, if signed build the signature block"""

    #print("hashListOrBuildSignatureBlock") # DBGDBG

    if(thisPdefList.SigAlgorithm in DEFINES.TPM_ALG_SIGN.values()):
      #if(thisPdefList.SigAlgorithm == DEFINES.LCP_POLSALG_NONE):
      if(thisPdefList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['NULL']):
        if(not self.hashThisList(pdef, thisPdefList, thisPolicyList, listCnt)):
           return False
      #elif(thisPdefList.SigAlgorithm == DEFINES.LCP_POLSALG_RSA_PKCS_15):
      #elif(thisPdefList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['RSASSA']):
      else:
        # Get the key from the file again in case it was replaced
        type = DEFINES.KEY_FILE_TYPE['PUBLIC_RSASSA']
        # once the entire file name is entered, verify it, else clear it
        if(thisPdefList.PubKeyFile.endswith(".pem")):
          if thisPdefList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['RSASSA']:
            type = DEFINES.KEY_FILE_TYPE['PUBLIC_RSASSA']
          elif thisPdefList.SigAlgorithm == DEFINES.TPM_ALG_SIGN['ECDSA']:
            type = DEFINES.KEY_FILE_TYPE['PUBLIC_ECDSA']
        pubkeyfile = os.path.join(self.dirname, thisPdefList.PubKeyFile)
        if not utilities.verifyKeyFile(pubkeyfile, type, thisPdefList):
          print("Invalid Public Key file %s" %(thisPdefList.PubKeyFile))
          return False
        self.createSignatureBlock(pdef, thisPdefList, thisPolicyList, listCnt)
    else:
      self.StatusBar.SetStatusText("hashListOrBuildSignatureBlock: aborting unknown pdef.SigAlgorithm=%x" % (thisPdefList.SigAlgorithm))
      return False

    return True

  def hashThisList(self, pdef, thisPdefList, thisPolicyList, listCnt):
    """hashThisList - when building an unsigned list, calcuate its hash"""
    func = 'hashThisList'

    print("%s: list %d of %d" % (func, listCnt+1, pdef.NumLists))

    # for each list in PolicyLists[]
    #   hash the list header
    #   hash the list's PolicyElements[]
    #   copy the hash to ListMeasurements[i]
    # Note: hash of the 1st list does NOT include the hash of the LCP_POLICY_DATA header

    try:
      hashAlgString = (key for key,val in DEFINES.TPM_ALG_HASH.items() if val == pdef.HashAlg).next()
    except StopIteration:
      self.StatusBar.SetStatusText("Hash Algorithm %d is not supported." % (pdef.HashAlg))
      print("%s: HashAlg %d is not supported" % (func, pdef.HashAlg))
      return False

    if hashAlgString == 'NULL':
      # Nothing to do
      return True
    else:
      hash = M2Crypto.EVP.MessageDigest(hashAlgString.lower())

    #lcpPolicyListHeader = self.packLcpPolicyListHeader(thisPolicyList)
    lcpPolicyListHeader = thisPolicyList.pack()
    hash.update(lcpPolicyListHeader)

    # hash this list's PolicyElements[], if they exist
    # Note: if list has no elements just hash the PolicyList header
    i=0
    numElements = len(thisPolicyList.PolicyElements)
    while(i < numElements):
      print("%s - checking if PolicyElement %d of %d exists" % (func, i, numElements)) # DBGDBG
      if(thisPolicyList.PolicyElements[i] != None):
        print("%s - packing PolicyElement %d of %d" % (func, i, numElements)) # DBGDBG
        #lcpPolicyListElement = self.packLcpPolicyElement(pdef, str(listCnt), i, thisPolicyList.PolicyElements[i])
        lcpPolicyListElement = thisPolicyList.PolicyElements[i].pack()
        hash.update(lcpPolicyListElement)
      else:
        break             # done after all existing elements are packed
      i += 1

    ListMeasurements.hashes[str(listCnt)] = hash.digest()
    #print("hashThisList - ListMeasurements[%d] hash size=%d hexdigest=%s" %
    #         (listCnt, hash.digest_size, hash.hexdigest()))   # DBGDBG
    return True

  def createSignatureBlock(self, pdef, thisPdefList, thisPolicyList, listCnt):
    """createSignatureBlock - when building a signed list, form its signature block"""

    try:
      signAlgStr = (key for key,val in DEFINES.TPM_ALG_SIGN.items() if val == thisPolicyList.SigAlgorithm).next()
    except StopIteration:
      print ("ERROR, Signing algorithm not supported (%d)" %(thisPolicyList.SigAlgorithm))
      return

    if 'RSA' in signAlgStr:
      # Create signature block for RSA signature structure
      thisPolicyList.Signature = LCP_RSA_SIGNATURE()
      #  Copy public key to Policy data structure
      thisPolicyList.Signature.PubkeyValue = thisPdefList.PubKeyData
    else:
      # Create signature block for ECC signature structure
      thisPolicyList.Signature = LCP_ECC_SIGNATURE()
      #  Copy public key to Policy data structure
      thisPolicyList.Signature.Qx = thisPdefList.PubKeyQx
      thisPolicyList.Signature.Qy = thisPdefList.PubKeyQy

    # Copy other common parameters

    #  RevocationCounter = PDEF.PolListInfo[i].RevocationCounter
    #  Note: Code differs from the tool spec algorithm in section 5.3.1 p19 as follows:
    #       - List Revocation Counters are incremented on the 1st change after a build, not here
    #       - RevokeCounter is sync'd to Revocation counter in list.onBuildButtonClick() not here
    thisPolicyList.Signature.RevocationCounter = thisPdefList.RevocationCounter

    #  PubkeySize = PDEF.PolListInfo[i].PubkeySize
    thisPolicyList.Signature.PubkeySize = int(thisPdefList.KeySize)
    print("createSignatureBlock - RevocationCounter=%d, keySize=%d per PDEF KeySize=%s" %
          (thisPolicyList.Signature.RevocationCounter, thisPolicyList.Signature.PubkeySize, thisPdefList.KeySize))       # DBGDBG

    #  Calculate Hash of public key using PDEF,HashAlg
    #  Copy to ListMeasurements[i]

    try:
      # reverse lookup of the hash algorithm name(key) for the given HashAlg value
      hashAlgStr = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == pdef.HashAlg)).next()
    except StopIteration:
      self.StatusBar.SetStatusText("Hash Algorithm is not supported, (pdef.HashAlg=%d)" % (pdef.HashAlg))
      return

    hash = M2Crypto.EVP.MessageDigest(hashAlgStr.lower())
    if 'RSA' in signAlgStr:
      hash.update(thisPolicyList.Signature.PubkeyValue)
    else:
      hash.update(thisPolicyList.Signature.Qx)
      hash.update(thisPolicyList.Signature.Qy)
    digest = hash.digest()
    ListMeasurements.hashes[str(listCnt)] = digest

    #  Build SigBlock[PubkeySize]
    #    Calculate hash of LIST using PDEF,HashAlg
    #    Encrypt hash using private key [ie Sign it]
    #       A.2.1 - For a signed list, the RSA signature is calculated over the entire LCP_POLICY_LIST struct,
    #         including the Signature member, EXCEPT for the SigBlock
    #    Copy encrypted hash to policy data structure LCP_SIGNATURE.SigBlock
    #  Done with signature

        # NOTE: The list is signed later, and the signature copied to LCP_SIGNATURE.SigBlock
        #       when the list is written to its .dat file in buildLcpPolicyDataFile()
        #       since that is when the 'signable' binary data is packed
        #       [vs. the Python object form of the data available here]


  # the last function in the file doesn't show up in the scope list in Understand for some reason!
  def stub(self):
    pass



#!/usr/bin/python
#  Copyright (c) 2013, Intel Corporation. All rights reserved.

# TXT Policy Generator Tool
# LcpPolicy and LcpPolicyData Classes - LCP_POLICY and LCP_POLICY_DATA File Structure
#

from defines import DEFINES
from ElementBase import *


class LCP_POLICY2( object ):
  """ LCP_POLICY2 Class"""

  #
  # LCP_POLICY2 struct
  #
  def __init__( self ):
    """__init__() - LCP_POLICY2 class constructor"""
    #self.Version                   # UINT16 - from pdef.PolVersion = 0x0300
    self.VersionMajor = 03          # UINT8
    self.VersionMinor = 00          # UINT8
    self.HashAlg = 00               # UINT16 - TPM_ALG_* from pdef.HashAlg
    self.PolicyType = 00            # UINT8 - 0=LIST, 1=ANY  - from pdef.PolicyType
    self.SINITMinVersion = 00       # UINT8 - from pdef.SinitMinVersion
    self.DataRevocationCounters = [0,0,0,0,0,0,0,0]  # UINT16  DataRevocationCounters[MAX_LISTS] Default is 0's
                                                     #  from pdef.DataRevocationCounters[]
    self.PolicyControl =  0         # UINT32  Encoding of (NPW, PCR17, Force PO), from pdef.PolicyControl
    self.MaxSinitMinVer = 0         # UINT8  PO only, reserved for PS
    self.MaxBiosAcMinVer = 0        # UINT8  PO only, reserved for PS
    self.LcpHashAlgMask = 0         # UINT16 HashMask for LCP eval.
    self.LcpSignAlgMask = 0
    self.AuxHashAlgMask = 0         # UINT16 Hash Mask for Auto-Promotion
    # Reserved                      # UINT16
    self.PolicyHash = []
    
    # Doesn't really need to use 2 variable for each hash size for python
    #self.PolicyHashSha1       = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0] # LCP_HASH
    #self.PolicyHashSha256     = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]

    #
    # The raw [ie binary] format LCP_POLICY.pol file is generated using struct.pack
    # based on the format string below.
    #
    # 'struct' module format string for a LCP_POLICY Structure
    #     where: B = UINT8, H = UINT16, L = UINT32
    #           '<' means Little Endian, std size, no alignment between fields
    #self.LcpPolicyFormatString       = "<BBHBB8HLBBHHH20B"   # for SHA1 policy hash
    #self.LcpPolicyFormatStringSha256 = "<BBHBB8HLBBHHH32B"   # for SHA256 policy hash
    self.LcpPolicyFormatStringNoHash = "<BBHBB8HLBBHLHH"
    self.LcpPolicyFormatString       = "<BBHBB8HLBBHHH"      # for policy hash - Hash added separately


class LCP_POLICY_DATA2( object):
  """LCP_POLICY_DATA2 class"""

  def __init__(self):
    """__init__() - LCP_POLICY_DATA2 class constructor"""

    self.FileSignature = "Intel(R) TXT LCP_POLICY_DATA\0\0\0\0"    # char [32]
    #self.Reserved[3]             # UINT8 [3]
    self.NumLists = 0             # UINT8
                                  # LCP_POLICY_LIST PolicyLists[NumLists];
    self.PolicyLists = {'0': LCP_POLICY_LIST2(), '1': LCP_POLICY_LIST2(),
                        '2': LCP_POLICY_LIST2(), '3': LCP_POLICY_LIST2(),
                        '4': LCP_POLICY_LIST2(), '5': LCP_POLICY_LIST2(),
                        '6': LCP_POLICY_LIST2(), '7': LCP_POLICY_LIST2()}

    # 'struct' module format string for a LCP_POLICY_DATA2 Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.HeaderFormatString       = "<32c3BB"   # same as LCP_POLICY_DATA header above

  def pack(self):
    """packLcpPolicyDataHeader"""

    #print("packLcpPolicyDataHeader") # DBGDBG
    #
    # LCP_POLICY_DATA Header
    #     CHAR8 FileSignature[32];
    #     UINT8 Reserved[3];
    #     UINT8 NumLists;
    return( pack(self.HeaderFormatString,
                self.FileSignature[0],self.FileSignature[1],
                self.FileSignature[2],self.FileSignature[3],
                self.FileSignature[4],self.FileSignature[5],
                self.FileSignature[6],self.FileSignature[7],
                self.FileSignature[8],self.FileSignature[9],
                self.FileSignature[10],self.FileSignature[11],
                self.FileSignature[12],self.FileSignature[13],
                self.FileSignature[14],self.FileSignature[15],
                self.FileSignature[16],self.FileSignature[17],
                self.FileSignature[18],self.FileSignature[19],
                self.FileSignature[20],self.FileSignature[21],
                self.FileSignature[22],self.FileSignature[23],
                self.FileSignature[24],self.FileSignature[25],
                self.FileSignature[26],self.FileSignature[27],
                self.FileSignature[28],self.FileSignature[29],
                self.FileSignature[30],self.FileSignature[31],
                0,0,0,
                self.NumLists)
          )


class LCP_RSA_SIGNATURE( object ):
  """LCP_RSA_SIGNATURE class"""

  def __init__(self):
    """__init__() - LCP_SIGNATURE class constructor"""

    self.RevocationCounter    = 0                # UINT16
    self.PubkeySize           = 2048             # UINT16 - key sizes are 1024, 2048 and 3072 bits. Default=2048
    self.PubkeyValue          = ""               # UINT8 PubkeyValue[PubkeySize]
    self.SigBlock             = ""               # UINT8 SigBlock[PubkeySize];   - as a binary string

    self.LcpRsaSignatureHdrFormatString  = "<HH"          # just the 1st 2 fields, other's size depends on KeySize

  def pack(self):
    """packLcpRsaSignatureHeader"""

    return( pack(self.LcpRsaSignatureHdrFormatString,
                self.RevocationCounter, self.PubkeySize/8)
            )


class LCP_ECC_SIGNATURE( object ):
  """LCP_SIGNATURE2 class"""

  def __init__(self):
    """__init__() - LCP_SIGNATURE class constructor"""

    self.RevocationCounter    = 0                # UINT16
    self.PubkeySize           = 256              # UINT16
    self.Reserved             = 0                # UINT32
    self.Qx                   = ""               # UINT8 x coordinate Public Key [PubkeySize]
    self.Qy                   = ""               # UINT8 y coordinate Public Key [PubkeySize]
    self.r                    = ""               # UINT8 r component of Signature
    self.s                    = ""               # UINT8 y component of Signature

    self.LcpEccSignatureHdrFormatString  = "<HHL"           # just the 1st 2 fields, other's size depends on KeySize

  def pack(self):
    """packLcpEccSignatureHeader - pack this LCP_ECC_SIGNATURE object except its SigBlock member and return the packed data"""

    # Note - The public key size in the pdef is the number of bits (1024, 2048, 3072)
    #        while LCP_SIGNATURE.PublicKeySize is the number if bytes, hence the /8 below
    return( pack(self.LcpEccSignatureHdrFormatString,
                self.RevocationCounter, self.PubkeySize/8, self.Reserved)
            )


class LCP_SIGNATURE2( object ):
  """LCP_SIGNATURE2 class"""

  def __init__(self):
    """__init__() - LCP_SIGNATURE class constructor"""

    self.RevocationCounter    = 0                # UINT16
    self.PubkeySize           = 2048             # UINT16 - key sizes are 1024, 2048 and 3072 bits. Default=2048
    self.PubkeyValue          = ""               # UINT8 PubkeyValue[PubkeySize]
    self.SigBlock             = ""               # UINT8 SigBlock[PubkeySize];   - as a binary string

    self.LcpSignatureHdrFormatString  = "<HH"             # just the 1st 2 fields, other's size depends on KeySize


class LCP_POLICY_LIST2( object ):
  """LCP_POLICY_LIST2 class"""

  def __init__(self):
    """__init__() - LCP_POLICY_LIST class constructor"""

    #self.Version             = 0x0200      # UINT16 - version 2.0
    self.VersionMajor        = 02
    self.VersionMinor        = 01
    self.SigAlgorithm        = 0x10         # UINT16 - one of LCP_POLSALG_*
    self.PolicyElementsSize  = 0            # UINT32
    # There can be 1 element of each type of each hash algorithm
    self.PolicyElements = []  # LCP_POLICY_ELEMENT
    #self.PolicyElements = {0:None, 1:None, 2:None, 3:None,  # LCP_POLICY_ELEMENT
    #                       4:None, 5:None, 6:None, 7:None}
    self.Signature = None                   # LCP_SIGNATURE - either LCP_RSA_SIGNATURE or LCP_ECC_SIGNATURE
                                            #  only included if SigAlgorithm != LCP_POLSALG_NONE

    # 'struct' module format string for a LCP_POLICY_LIST2 Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.HeaderFormatString       = "<BBHL"

  def pack(self):
    """packLcpPolicyListHeader"""

    #print("packLcpPolicyListHeader") # DBGDBG
    # LCP_POLICY_LIST Header
    #     UINT16 Version;                         // 0x0100                   Header
    #     UINT16 SigAlgorithm;                    // one of LCP_POLSALG_*     Header
    #     UINT32 PolicyElementsSize;              //                          Header
    return( pack(self.HeaderFormatString,
                self.VersionMinor, self.VersionMajor,
                self.SigAlgorithm, self.PolicyElementsSize)
            )

#
# LCP_MLE_ELEMENT2
#
class LCP_MLE_ELEMENT2( ElementBase ):
  """LCP_MLE_ELEMENT2 class"""

  def __init__(self):
    """__init__() - LCP_MLE_ELEMENT class constructor"""

    self.ElementSize      = 0                       # UINT32 - header
    self.ElementType      = DEFINES.LCP_POLELT_TYPE_MLE2     # UINT32 - header
    self.PolEltControl    = 0                       # UINT32 - header
    self.SINITMinVersion  = 0                       # UINT8
    # Reserved                                      # UINT8
    self.HashAlg          = 0                       # UINT16
    self.NumHashes        = 0                       # UINT16
    self.Hashes           = []                      # LCP_HASH Hashes[NumHashes]

    # 'struct' module format string for a LCP_MLE_ELEMENT2 Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.MleDataFormatString              = "<LLLBBHH"
    self.MleDataSha1HashFormatString      = "<20B"
    self.MleDataSha256HashFormatString    = "<32B"

  def pack(self):
    """packLcpPolicyMleElement - pack this list's MLE element """
    func = 'packLcpPolicyMleElement'

    print("%s packing MLE hdr" % (func))   # DBGDBG
    
    # Initialize to None to check for supported HashAlg.
    hashAlgName = None
    # reverse lookup of the hash algorithm name(key) for the given HashAlg value
    hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == self.HashAlg)).next()
    if (hashAlgName == None):
      print ("MLE elements with unsupported hash algorithm, aborting build")
      print("%s - build failed, see status bar" % (func))  # DBGDBG
      return

    # pack the element based on its type and return the binary string
    elementData = pack(self.MleDataFormatString,
        self.ElementSize, self.ElementType,
        self.PolEltControl, self.SINITMinVersion, 0,
        self.HashAlg, self.NumHashes)
        
    print("%s PolEltControl=%d, SINITMinVersion=%d" %(func, self.PolEltControl, self.SINITMinVersion))   # DBGDBG
    fileCnt = 0
    while(fileCnt < self.NumHashes):
      print("%s packing MLE hash %d" % (func, fileCnt))   # DBGDBG
      elementData += self.packHash(hashAlgName, self.Hashes[fileCnt])
      fileCnt += 1

    return( elementData )

#
# LCP_STM_ELEMENT2
#
class LCP_STM_ELEMENT2( ElementBase ):
  """LCP_STM_ELEMENT2 class"""

  def __init__(self):
    """__init__() - LCP_STM_ELEMENT class constructor"""

    self.ElementSize      = 0                       # UINT32 - header
    self.ElementType      = DEFINES.LCP_POLELT_TYPE_STM2     # UINT32 - header
    self.PolEltControl    = 0                       # UINT32 - header
    self.HashAlg          = 0                       # UINT16
    self.NumHashes        = 0                       # UINT16
    self.Hashes           = []                      # LCP_HASH Hashes[NumHashes]

    # 'struct' module format string for a LCP_STM_ELEMENT2 Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.StmDataFormatString              = "<LLLHH"
    self.StmDataSha1HashFormatString      = "<20B"
    self.StmDataSha256HashFormatString    = "<32B"

  def pack(self):
    """packLcpPolicyStmElement - pack this list's STM element """
    func = 'packLcpPolicyStmElement'

    print("%s packing STM hdr" % (func))   # DBGDBG

    # Initialize to None to check for supported HashAlg.
    hashAlgName = None
    # reverse lookup of the hash algorithm name(key) for the given HashAlg value
    hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == self.HashAlg)).next()
    if (hashAlgName == None):
      print ("STM elements with unsupported hash algorithm, aborting build")
      print("%s - build failed, see status bar" % (func))  # DBGDBG
      return

    # pack the element based on its type and return the binary string
    elementData = pack(self.StmDataFormatString,
        self.ElementSize, self.ElementType, self.PolEltControl,
        self.HashAlg, self.NumHashes)
        
    print("%s PolEltControl=%d" %(func, self.PolEltControl))   # DBGDBG
    fileCnt = 0
    while(fileCnt < self.NumHashes):
      print("%s packing STM hash %d" % (func, fileCnt))   # DBGDBG
      elementData += self.packHash(hashAlgName, self.Hashes[fileCnt])
      fileCnt += 1

    return( elementData )

#
# LCP_PCONF_ELEMENT2
#
class LCP_PCONF_ELEMENT2( ElementBase ):
  """LCP_PCONF_ELEMENT2 class"""

  def __init__(self):
    """__init__() - LCP_PCONF_ELEMENT class constructor"""

    self.ElementSize      = 0                             # UINT32 - header
    self.ElementType      = DEFINES.LCP_POLELT_TYPE_PCONF2 # UINT32 - header
    self.PolEltControl    = 0                             # UINT32 - header
    self.HashAlg          = 0                             # UINT16
    self.NumPCRInfos      = 0                             # UINT16
    self.PCRInfos         = []                            # TPMS_QUOTE_INFO PCRInfos[NumPCRInfos]

    # 'struct' module format string for a LCP_PCONF_ELEMENT2 Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.PconfDataFormatString        = "<LLLHH"    # ElementSize,Type,Ctl,HashAlg,NumPcrInfos
    
    self.PconfDataSha1FormatString   = "<20B"
    self.PconfDataSha256FormatString = "<32B"

  def pack(self):
    """packLcpPolicyPconfElement - pack this list's PCONF element """

    func = "packLcpPolicyPconfElement"
    print("%s packing PCONF hdr" % (func))   # DBGDBG
    
    # Initialize to None to check for supported HashAlg.
    hashAlgName = None
    # reverse lookup of the hash algorithm name(key) for the given HashAlg value
    hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == self.HashAlg)).next()
    if (hashAlgName == None):
      print ("PCONF elements with unsupported hash algorithm, aborting build")
      print("%s - build failed, see status bar" % (func))  # DBGDBG
      return
      
    # pack the element based on its type and return the binary string
    elementData = pack(self.PconfDataFormatString,
        self.ElementSize, self.ElementType,
        self.PolEltControl, self.HashAlg,
        self.NumPCRInfos)

    numPcrInfos = 0
    while(numPcrInfos < self.NumPCRInfos):
      print("%s packing Pconf PcrInfo %d" %(func, numPcrInfos))   # DBGDBG
      elementData += self.PCRInfos[numPcrInfos].pack()
      numPcrInfos += 1
      # end of numPcrInfo while

    return( elementData )


#
# TPMS_QUOTE_INFO
#
class TPMS_QUOTE_INFO( ElementBase ):
  """TPMS_QUOTE_INFO class"""

  def __init__(self):
    """__init__() - TPMS_QUOTE_INFO class constructor"""

    self.pcrSelect         = TPML_PCR_SELECTION()      # TPML_PCR_SELECTION	pcrSelect
    self.pcrDigest         = TPM2B_DIGEST()           # UINT8
    
    # 'struct' module format string for a TPMS_QUOTE_INFO Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.TpmsQuoteInfoFormatString = ">LHB3BH"   # TPMS_QUOTE_INFO: count,hash,sizeOfSelect,pcrSelect[],bufferSize
    
  def pack(self):
    # TPMS_QUOTE_INFO is big endian so packing TPML_PCR_SELECTION and TPM2B_DIGEST here,
    # in case those can be used elsewhere as little endian.

    func = "packLcpPolicyPconfElement_tpms_quote_info"
    thisTpmlPcrSelection = self.pcrSelect
    thisTpm2bDigest      = self.pcrDigest
    thisTpmsPcrSelection = thisTpmlPcrSelection.pcrSelections
    thisBuf              = thisTpm2bDigest.buffer
    #print("packLcpPolicyPconfElement TypeOf: thisPcrInfo.digestAtRelease[0]=%s, _GlobalPcrHash=%s, _GlobalPcrHash[0]=%s" %
    #      (type(thisPcrInfo.digestAtRelease[0]), type(_GlobalPcrHash), type(_GlobalPcrHash[0])))  # DBGDBG

    elementData = pack(self.TpmsQuoteInfoFormatString,
        thisTpmlPcrSelection.count,
        thisTpmsPcrSelection.hash, thisTpmsPcrSelection.sizeOfSelect,
        thisTpmsPcrSelection.pcrSelect[0],
        thisTpmsPcrSelection.pcrSelect[1],
        thisTpmsPcrSelection.pcrSelect[2],
        thisTpm2bDigest.size)

    # pack TPM2B_DIGEST.buffer. this does not verify the size as multiple of hash digest size.
    b = bytes()
    elementData += b.join(pack('B', val) for val in thisBuf)
    
    return( elementData )


#
# TPML_PCR_SELECTION
#
class TPML_PCR_SELECTION( object ):
  """TPML_PCR_SELECTION class"""

  def __init__(self):
    """__init__() - TPML_PCR_SELECTION class constructor"""

    self.count              = 0                       # UINT32	count;	# must be 1 for use in PCONF
    self.pcrSelections      = TPMS_PCR_SELECTION()    # TPMS_PCR_SELECTION	pcrSelections;


#
# TPMS_PCR_SELECTION
#
class  TPMS_PCR_SELECTION( object ):
  """TPMS_PCR_SELECTION class"""

  def __init__(self):
    """__init__() - TPMS_PCR_SELECTION class constructor"""

    self.hash               = 0
    self.sizeOfSelect       = 0x03                    # UINT8	sizeofSelect;
    self.pcrSelect          = [0, 0, 0]               # UINT8 pcrSelect[sizeofSelect];


#
# TPM2B_DIGEST
#
class TPM2B_DIGEST( object ):
  """TPM2B_DIGEST class"""

  def __init__(self):
    """__init__() - TPM2B_DIGEST class constructor"""

    self.size              = 0      # size of buffer  # UINT16
    self.buffer            = []                       # UINT8	buffer[size];
    
    # 'struct' module format string for a TPMS_QUOTE_INFO Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.Tpm2BDigestFormatString = ">H"

  def pack(self):
    elementData = pack(self.Tpm2BDigestFormatString,   # TODO: is this always big endian?
                      self.size)
    b = bytes()
    elementData += b.join(pack('B', val) for val in self.buffer)
  
    return elementData


#
# LCP_SBIOS_ELEMENT2
#
class LCP_SBIOS_ELEMENT2( ElementBase ):
  """LCP_SBIOS_ELEMENT2 class"""

  def __init__(self):
    """__init__() - LCP_SBIOS_ELEMENT class constructor"""

    self.ElementSize    = 0                            # UINT32 - header
    self.ElementType    = DEFINES.LCP_POLELT_TYPE_SBIOS2        # UINT32 - header
    self.PolEltControl  = 0                            # UINT32 - header
    self.HashAlg        = 0                            # UINT16
    #self.Reserved1[2]                                 # UINT8 0, 0
    self.FallbackHash   = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
                                                       # LCP_HASH
    #self.Reserved2                                    # UINT16 0x0000
    self.NumHashes      = 0                            # UINT16
    self.Hashes         = []                           # LCP_HASH Hashes[NumHashes]

    # 'struct' module format string for a LCP_SBIOS_ELEMENT2 Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.SbiosDataFormatString           = "<LLLHBB"
    self.SbiosDataHashFormatString       = "<HH"
    self.SbiosDataSha1FormatString       = "<LLLHBB20BHH"       # SHA1   fallback hash
    self.SbiosDataSha256FormatString     = "<LLLHBB32BHH"       # SHA256 fallback hash
    self.SbiosDataSha1HashFormatString   = "<20B"
    self.SbiosDataSha256HashFormatString = "<32B"

  def pack(self):
    """packLcpPolicySbiosElement - pack this list's SBIOS element """

    func = "packLcpPolicySbiosElement"
    print("%s packing SBIOS hdr" % (func))   # DBGDBG
    
    # Initialize to None to check for supported HashAlg.
    hashAlgName = None
    # reverse lookup of the hash algorithm name(key) for the given HashAlg value
    hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == self.HashAlg)).next()
    if (hashAlgName == None):
      print ("SBIOS elements with unsupported hash algorithm, aborting build")
      print("%s - build failed, see status bar" % (func))  # DBGDBG
      return

    elementData = pack(self.SbiosDataFormatString,
        self.ElementSize, self.ElementType, self.PolEltControl,
        self.HashAlg, 0,0)
        
    # pack the element based on its type and return the binary string
    elementData += self.packHash(hashAlgName, self.FallbackHash)
    
    elementData += pack(self.SbiosDataHashFormatString,
        0, self.NumHashes)

    fileCnt = 0
    while(fileCnt < self.NumHashes):
      print("%s packing SBIOS hash %d" % (func, fileCnt))   # DBGDBG
      elementData += self.packHash(hashAlgName, self.Hashes[fileCnt])
      fileCnt += 1

    return( elementData )

    
#
#   UINT32 ElementSize;
#   UINT32 ElementType;
#   UINT32 PolEltControl;
#
#   #define LCP_POLELT_TYPE_MLE   0
#   #define LCP_POLELT_TYPE_PCONF 1
#   #define LCP_POLELT_TYPE_SBIOS 2
#
#   typedef struct {
#     UINT32 ElementSize;
#     UINT32 ElementType;       // Type = 0 = LCP_POLELT_TYPE_MLE
#     UINT32 PolEltControl;
#     UINT8 SINITMinVersion;
#     UINT8 HashAlg;          // one of LCP_POLHALG_*
#     UINT16 NumHashes;
#     LCP_HASH Hashes[NumHashes];
#   } LCP_MLE_ELEMENT;
#
class LCP_MLE_ELEMENT( ElementBase ):
  """LCP_MLE_ELEMENT class"""

  def __init__(self):
    """__init__() - LCP_MLE_ELEMENT class constructor"""

    self.ElementSize      = 0                           # UINT32 - header
    self.ElementType      = DEFINES.LCP_POLELT_TYPE_MLE # UINT32 - header
    self.PolEltControl    = 0                           # UINT32 - header
    self.SINITMinVersion  = 0                           # UINT8
    self.HashAlg          = 0                           # UINT8
    self.NumHashes        = 0                           # UINT8
    self.Hashes           = []                          # LCP_HASH Hashes[NumHashes]

    # 'struct' module format string for a LCP_MLE_ELEMENT Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.MleDataFormatString              = "<LLLBBH"
    self.MleDataSha1HashFormatString      = "<20B"
    self.MleDataSha256HashFormatString    = "<32B"

  def pack(self):
    """packLcpPolicyMleLegacyElement - pack this list's MLE Legacy element """

    func = "packLcpPolicyMleLegacyElement"
    print("%s packing MLE hdr" % (func))   # DBGDBG

    # Initialize to None to check for supported HashAlg.
    hashAlgName = None
    try:
      # reverse lookup of the hash algorithm name(key) for the given HashAlg value
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == self.HashAlg)).next()
      hashAlgName = 'SHA1'
    except StopIteration:
      print ("MLE elements with unsupported hash algorithm, aborting build")
      print("%s - build failed, see status bar" % (func))  # DBGDBG
      return
      
    if(self.HashAlg != DEFINES.TPM_ALG_HASH['SHA1_LEGACY']):
      #elementFileFormat = MleDataSha256HashFormatString
      print ("MLE Legacy elements only support SHA1, aborting build")
      return

    # pack the element based on its type and return the binary string
    elementData = pack( self.MleDataFormatString,
        self.ElementSize, self.ElementType,
        self.PolEltControl, self.SINITMinVersion,
        self.HashAlg, self.NumHashes)
    print("%s PolEltControl=%d, SINITMinVersion=%d" %(func, self.PolEltControl, self.SINITMinVersion))   # DBGDBG
    
    fileCnt = 0
    while(fileCnt < self.NumHashes):
      print("%s packing MLE hash %d" % (func, fileCnt))   # DBGDBG
      elementData += self.packHash(hashAlgName, self.Hashes[fileCnt])
      fileCnt += 1

    return( elementData )


#   typedef struct {
#     UINT32 ElementSize;
#     UINT32 ElementType;       // Type = 1 = LCP_POLELT_TYPE_PCONF
#     UINT32 PolEltControl;
#     UINT16 NumPCRInfos;       // Can be 0
#     TPM_PCR_INFO_SHORT PCRInfos[NumPCRInfos];
#   } LCP_PCONF_ELEMENT;
#
class LCP_PCONF_ELEMENT( ElementBase ):
  """LCP_PCONF_ELEMENT class"""

  def __init__(self):
    """__init__() - LCP_PCONF_ELEMENT class constructor"""

    self.ElementSize      = 0                             # UINT32 - header
    self.ElementType      = DEFINES.LCP_POLELT_TYPE_PCONF # UINT32 - header
    self.PolEltControl    = 0                             # UINT32 - header
    self.NumPCRInfos      = 0                             # UINT16
    self.PCRInfos         = []                            # TPM_PCR_INFO_SHORT PCRInfos[NumPCRInfos]

    # 'struct' module format string for a LCP_PCONF_ELEMENT Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    self.PconfDataFormatString            = "<LLLH"    # Size,Type,Ctl,NumPcrInfos
    self.PconfDataSha1PcrInfoFormatString = ">H3BB"    # TPM_PCR_INFO_SHORT: sizeOfSelect,pcrSelect,localityAtRelease
    # Note: the 20B of the digestAtRelease is output directly, ie does not need to be packed
    # since all ready a binary string

  def pack(self):
    """packLcpPolicyPconfLegacyElement - pack this list's PCONF Legacy element """

    func = "packLcpPolicyPconfLegacyElement"
    print("%s packing PCONF hdr" %(func))   # DBGDBG

    #Note:  LCP_PCONF_ELEMENT has no HashAlg member. If HashAlg info is needed, check HashAlg using PDEF's PCONF_INFO

    # pack the element based on its type and return the binary string
    elementData = pack( self.PconfDataFormatString,
                        self.ElementSize, self.ElementType,
                        self.PolEltControl, self.NumPCRInfos)

    i = 0
    while(i < self.NumPCRInfos):
      thisPcrInfo = self.PCRInfos[i]
      print("%s packing Pconf PcrInfo %d" %(func, i))   # DBGDBG
      #print("packLcpPolicyPconfLegacyElement TypeOf: thisPcrInfo.digestAtRelease[0]=%s, _GlobalPcrHash=%s, _GlobalPcrHash[0]=%s" %
      #      (type(thisPcrInfo.digestAtRelease[0]), type(_GlobalPcrHash), type(_GlobalPcrHash[0])))  # DBGDBG


      # Note: SizeOfSelect field is byteswapped  by WinLCPTool
      # see Win LCP Apps\plugins\pconf_elt.cpp line 216:
      #   pcr_info->pcr_selection.size_of_select = 3;
      #   // TPM structures are big-endian, so byte-swap
      #   bswap16(&pcr_info->pcr_selection.size_of_select);
      #
      # So do likewise

      elementData += pack( self.PconfDataSha1PcrInfoFormatString,
                           thisPcrInfo.pcrSelection.sizeOfSelect,
                           thisPcrInfo.pcrSelection.pcrSelect[0],
                           thisPcrInfo.pcrSelection.pcrSelect[1],
                           thisPcrInfo.pcrSelection.pcrSelect[2],
                           thisPcrInfo.localityAtRelease)

      # thisPcrInfo.digestAtRelease is the binary string produced by hash.digest()
      # struct.pack() requires ints, not strings.
      # But don't need to pack it, just output it directly
      elementData += thisPcrInfo.digestAtRelease
      i += 1

    return( elementData )


#   #define TPM_LOCALITY_SELECTION BYTE
#
#   typedef struct tdTPM_PCR_INFO_SHORT{
#     TPM_PCR_SELECTION pcrSelection;
#     UINT8 localityAtRelease;     //0x1F
#     TPM_COMPOSITE_HASH digestAtRelease;
#   } TPM_PCR_INFO_SHORT;
class TPM_PCR_INFO_SHORT( object ):
  """TPM_PCR_INFO_SHORT class"""

  def __init__(self):
    """__init__() - TPM_PCR_INFO_SHORT class constructor"""

    self.pcrSelection         = TPM_PCR_SELECTION()   # TPM_PCR_SELECTION
    self.localityAtRelease    = 0x1f                  # UINT8
    self.digestAtRelease      = []


#
#   typedef struct tdTPM_PCR_SELECTION {
#     UINT16 sizeOfSelect;        // shall be 0x0003
#     [size_is(sizeOfSelect)] BYTE pcrSelect[];
#   } TPM_PCR_SELECTION;
#
class TPM_PCR_SELECTION( object ):
  """TPM_PCR_SELECTION class"""

  def __init__(self):
    """__init__() - TPM_PCR_SELECTION class constructor"""

    self.sizeOfSelect       = 0x0003                  # UINT16
    self.pcrSelect          = [0, 0, 0]               # [size_is(sizeOfSelect)] BYTE pcrSelect[]


#   typedef struct tdTPM_DIGEST{
#     BYTE digest[digestSize];    // digestSize = 20 bytes for SHA1
#   } TPM_DIGEST;
#
#   typedef TPM_DIGEST TPM_COMPOSITE_HASH; // hash of multiple measurements
#
class TPM_DIGEST( object ):
  """TPM_DIGEST class"""

  def __init__(self):
    """__init__() - TPM_DIGEST class constructor"""

    self.digest            = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                                 0,0,0,0]             # BYTE digest[digestSize] MAX_HASH = 20 for SHA1


#   typedef struct {
#     UINT32  ElementSize;
#     UINT32  ElementType;      // Type = 2 = LCP_POLELT_TYPE_SBIOS
#     UINT32  PolEltControl;
#     UINT8   HashAlg;          // one of LCP_POLHALG_*
#     UINT8   Reserved1[3];     // 0, 0, 0
#     LCP_HASH FallbackHash;
#     UINT16  Reserved2;        // 0x0000
#     UINT16  NumHashes;
#     LCP_HASH Hashes[NumHashes];
#   } LCP_SBIOS_ELEMENT
#
class LCP_SBIOS_ELEMENT( ElementBase ):
  """LCP_SBIOS_ELEMENT class"""

  def __init__(self):
    """__init__() - LCP_SBIOS_ELEMENT class constructor"""

    self.ElementSize    = 0                                # UINT32 - header
    self.ElementType    = DEFINES.LCP_POLELT_TYPE_SBIOS    # UINT32 - header
    self.PolEltControl  = 0                                # UINT32 - header
    self.HashAlg        = 0                                # UINT8
    #self.Reserved1[3]                                     # UINT8 0, 0, 0
    self.FallbackHash   = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
                                                           # LCP_HASH
    #self.Reserved2                                        # UINT16 0x0000
    self.NumHashes      = 0                                # UINT16
    self.Hashes         = []                               # LCP_HASH Hashes[NumHashes]

    # 'struct' module format string for a LCP_SBIOS_ELEMENT Structure
    #     where: B = UINT8, H = UINT16, L = UINT32 c = char
    #           '<' means Little Endian, std size, no alignment between fields
    #self.SbiosDataFormatString           = "<LLLB3BHH"
    self.SbiosDataFormatString           = "<LLLBBBB20BHH"
    self.SbiosDataSha1HashFormatString   = "<20B"
    self.SbiosDataSha256HashFormatString = "<32B"

  def pack(self):
    """packLcpPolicySbiosElement - pack this list's SBIOS element """

    func = "packLcpPolicySbiosElement"
    
    print("%s packing SBIOS hdr" %(func))   # DBGDBG

    # Initialize to None to check for supported HashAlg.
    hashAlgName = None
    try:
      # reverse lookup of the hash algorithm name(key) for the given HashAlg value
      hashAlgName = (key for key,val in DEFINES.TPM_ALG_HASH.items() if (val == self.HashAlg)).next()
      hashAlgName = 'SHA1'
    except StopIteration:
      print ("SBIOS Legacy elements only support SHA1, aborting build")
      print("%s - build failed, see status bar" % (func))  # DBGDBG
      return
      
    # SBIOS legacy element must be SHA1
    #if(self.HashAlg != DEFINES.TPM_ALG_HASH['SHA1_LEGACY']):
    #  print ("SBIOS Legacy elements only support SHA1, aborting build")
    #  return

    # pack the element based on its type and return the binary string
    elementData = pack(
        self.SbiosDataFormatString,
        self.ElementSize, self.ElementType, self.PolEltControl,
        self.HashAlg, 0,0,0,
        self.FallbackHash[0],  self.FallbackHash[1],  self.FallbackHash[2],  self.FallbackHash[3],  
        self.FallbackHash[4],  self.FallbackHash[5],  self.FallbackHash[6],  self.FallbackHash[7],
        self.FallbackHash[8],  self.FallbackHash[9],  self.FallbackHash[10], self.FallbackHash[11],
        self.FallbackHash[12], self.FallbackHash[13], self.FallbackHash[14], self.FallbackHash[15],
        self.FallbackHash[16], self.FallbackHash[17], self.FallbackHash[18], self.FallbackHash[19],
        0, self.NumHashes )
    fileCnt = 0
    while(fileCnt < self.NumHashes):
      elementData += self.packHash(hashAlgName, self.Hashes[fileCnt])
      print("%s packing SBIOS hash %d" % (func, fileCnt))   # DBGDBG
      fileCnt += 1

    return( elementData )



#
# DIGEST ListMeasurements[PDEF.NumLists]
#
class ListMeasurements( object ):
  """ListMeasurements class"""

  def __init__(self):
    """__init__() - ListMeasurements class constructor"""

    # SHA1 - 20 bytes
    self.hashes =   {'0': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                     '1': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                     '2': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                     '3': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                     '4': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                     '5': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                     '6': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                     '7': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}

    # SHA256 - 32 bytes
    #self.hashes32 = {'0': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    #                 '1': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    #                 '2': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    #                 '3': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    #                 '4': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    #                 '5': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    #                 '6': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    #                 '7': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}

